from functools import lru_cache

import shapely
from pyproj import CRS, Transformer


@lru_cache
def is_crs_always_xy(srid: int) -> bool:
    crs = CRS.from_user_input(srid)
    # Pseudo-Mercator WGS 84, always east, north
    crs_wgs84 = CRS.from_user_input(3857)

    transformer = Transformer.from_crs(crs, crs_wgs84)
    transformer_xy = Transformer.from_crs(crs, crs_wgs84, always_xy=True)

    transformed = transformer.transform(10, 20)
    transformer_xy = transformer_xy.transform(10, 20)

    return transformed == transformer_xy


@lru_cache
def get_crs_bounding_area(srid: int) -> shapely.geometry.Polygon:
    crs = CRS.from_user_input(srid)
    transformer = Transformer.from_crs(crs.geodetic_crs, crs, always_xy=True)
    west, south, east, north = transformer.transform_bounds(*crs.area_of_use.bounds)

    is_lon_lat = is_crs_always_xy(srid)

    if is_lon_lat:
        return shapely.geometry.box(
            minx=west,
            maxx=east,
            miny=south,
            maxy=north
        )

    return shapely.geometry.box(
        minx=south,
        maxx=north,
        miny=west,
        maxy=east
    )



from abc import ABC, abstractmethod
from sqlalchemy.dialects.postgresql import UUID
from typing import List, Tuple, Dict

import sqlalchemy as sa
from alembic.operations import Operations


class MigrationAction(ABC):
    @abstractmethod
    def execute(self, op: Operations):
        pass


class CreateTableMigrationAction(MigrationAction):
    def __init__(self, table_name: str, columns: List[sa.Column]):
        self.table_name = table_name
        self.columns = columns

    def execute(self, op: Operations):
        op.create_table(self.table_name, *self.columns)


class DropTableMigrationAction(MigrationAction):
    def __init__(self, table_name: str):
        self.table_name = table_name

    def execute(self, op: Operations):
        op.drop_table(table_name=self.table_name)


class RenameTableMigrationAction(MigrationAction):
    def __init__(self, old_table_name: str, new_table_name: str):
        self.old_table_name = old_table_name
        self.new_table_name = new_table_name

    def execute(self, op: Operations):
        op.rename_table(old_table_name=self.old_table_name, new_table_name=self.new_table_name)


class AddColumnMigrationAction(MigrationAction):
    def __init__(self, table_name: str, column: sa.Column):
        self.table_name = table_name
        self.column = column

    def execute(self, op: Operations):
        op.add_column(table_name=self.table_name, column=self.column)


class DropColumnMigrationAction(MigrationAction):
    def __init__(self, table_name: str, column_name: str):
        self.table_name = table_name
        self.column_name = column_name

    def execute(self, op: Operations):
        op.drop_column(table_name=self.table_name, column_name=self.column_name)


class AlterColumnMigrationAction(MigrationAction):
    def __init__(self, table_name: str, column_name: str, nullable: bool = None, new_column_name: str = None,
                 type_=None, using: str = None):
        self.table_name = table_name
        self.column_name = column_name
        self.nullable = nullable
        self.new_column_name = new_column_name
        self.type_ = type_
        self.using = using

    def execute(self, op: Operations):
        op.alter_column(
            table_name=self.table_name,
            column_name=self.column_name,
            nullable=self.nullable,
            new_column_name=self.new_column_name,
            type_=self.type_,
            postgresql_using=self.using
        )


class DropConstraintMigrationAction(MigrationAction):
    def __init__(self, table_name: str, constraint_name: str):
        self.table_name = table_name
        self.constraint_name = constraint_name

    def execute(self, op: Operations):
        op.drop_constraint(constraint_name=self.constraint_name, table_name=self.table_name)


class DropIndexMigrationAction(MigrationAction):
    def __init__(self, table_name: str, index_name: str):
        self.table_name = table_name
        self.index_name = index_name

    def execute(self, op: Operations):
        op.drop_index(index_name=self.index_name, table_name=self.table_name)


class CreateUniqueConstraintMigrationAction(MigrationAction):
    def __init__(self, table_name: str, constraint_name: str, columns: List[str]):
        self.table_name = table_name
        self.constraint_name = constraint_name
        self.columns = columns

    def execute(self, op: Operations):
        op.create_unique_constraint(constraint_name=self.constraint_name, table_name=self.table_name,
                                    columns=self.columns)


class RenameConstraintMigrationAction(MigrationAction):
    def __init__(self, table_name: str, old_constraint_name: str, new_constraint_name: str):
        self.query = f'ALTER TABLE "{table_name}" RENAME CONSTRAINT "{old_constraint_name}" TO "{new_constraint_name}"'

    def execute(self, op: Operations):
        replaced = self.query.replace(":", "\\:")
        op.execute(replaced)


class CreatePrimaryKeyMigrationAction(MigrationAction):
    def __init__(self, table_name: str, constraint_name: str, columns: List[str]):
        self.table_name = table_name
        self.constraint_name = constraint_name
        self.columns = columns

    def execute(self, op: Operations):
        op.create_primary_key(constraint_name=self.constraint_name, table_name=self.table_name, columns=self.columns)


class CreateIndexMigrationAction(MigrationAction):
    def __init__(self, table_name: str, index_name: str, columns: List[str], using: str = None):
        self.table_name = table_name
        self.index_name = index_name
        self.columns = columns
        self.using = using

    def execute(self, op: Operations):
        op.create_index(index_name=self.index_name, table_name=self.table_name, columns=self.columns,
                        postgresql_using=self.using)


class RenameIndexMigrationAction(MigrationAction):
    def __init__(self, old_index_name: str, new_index_name: str):
        self.query = f'ALTER INDEX "{old_index_name}" RENAME TO "{new_index_name}"'

    def execute(self, op: Operations):
        replaced = self.query.replace(":", "\\:")
        op.execute(replaced)


class DowngradeTransferDataMigrationAction(MigrationAction):
    """Migration Action used to convert columns into external ref columns.

    Args:
        table_name: Name of currently migrating table
        referenced_table_name: Name of referenced table (ref.model)
        source_column: Original column which is being converted to external ref (currently system only allows conversion from one column)
        columns: Mapping of new ref column names and referenced table's columns.
        target: Name of foreign table's mapping column

    Example:
        Let's say you have two tables called "Country" and "City", where the "Country" table references the "City" table.
        In the current scenario, the reference level is 4, meaning there is foreign key constraint, and it's mapped through "_id" column.

        The "City" table is referenced using these columns: "city._id".
        It's primary key consist of "id" and "code" columns.

        To convert this reference to level 3 or lower, you can use:

            DowngradeTransferDataMigrationAction(
                table_name="Country",
                referenced_table_name="City",
                ref_column=Column("city._id", "Country"),
                columns={
                    "city.id": Column("id", "City"),
                    "city.code": Column("code", "City")
                },
                target="_id"
            )
    """

    def __init__(
        self,
        table_name: str,
        referenced_table_name: str,
        source_column: sa.Column,
        columns: Dict[str, sa.Column],
        target: str
    ):
        original_table = sa.Table(
            table_name,
            sa.MetaData(),
            source_column._copy(),
            *[sa.Column(key, column.type) for key, column in columns.items()]
        )
        foreign_table = sa.Table(
            referenced_table_name,
            sa.MetaData(),
            sa.Column('_id', UUID),
            *[column._copy() for column in columns.values()]
        )
        self.query = original_table.update().values(
            **{
                key: foreign_table.columns[column.name]
                for key, column in columns.items()
            }
        ).where(
            original_table.columns[source_column.name] == foreign_table.columns[target]
        )

    def execute(self, op: Operations):
        op.execute(self.query)


class UpgradeTransferDataMigrationAction(MigrationAction):
    """Migration Action used to convert multiple columns into one internal ref column ("column._id").

    Args:
        table_name: Name of currently migrating table
        referenced_table_name: Name of referenced table (ref.model)
        ref_column: Newly created ref column, ending with "_id", ex: "column._id"
        columns: Dictionary of columns used to migrate, key is "referenced_table" column name. Value is migration table's column

    Example:
        Let's say you have two tables called "Country" and "City", where the "Country" table references the "City" table.
        In the current scenario, the reference level is 3, meaning there are no foreign keys and the reference is done using normal columns.

        The "City" table is referenced using these columns: "city.id", "city.code".

        To convert this reference to level 4 or higher, you can use:

            UpgradeTransferDataMigrationAction(
                table_name="Country",
                referenced_table_name="City",
                ref_column=Column("city._id"),
                columns={
                    "id": Column("city.id"),
                    "code": Column("city.code")
                }
            )
    """

    def __init__(
        self,
        table_name: str,
        referenced_table_name: str,
        ref_column: sa.Column,
        columns: Dict[str, sa.Column]
    ):
        original_table = sa.Table(
            table_name,
            sa.MetaData(),
            ref_column._copy(),
            *[column._copy() for column in columns.values()]
        )
        referenced_table = sa.Table(
            referenced_table_name,
            sa.MetaData(),
            sa.Column('_id', UUID),
            *[sa.Column(key, column.type) for key, column in columns.items()]
        )
        self.query = original_table.update().values(
            **{
                ref_column.name: referenced_table.columns['_id']
            }
        ).where(
            sa.and_(
                *[original_table.columns[column.name] == referenced_table.columns[key] for key, column in
                  columns.items()]
            )
        )

    def execute(self, op: Operations):
        op.execute(self.query)


class CreateForeignKeyMigrationAction(MigrationAction):
    def __init__(self, source_table: str, referent_table: str, constraint_name: str, local_cols: list,
                 remote_cols: list):
        self.constraint_name = constraint_name
        self.source_table = source_table
        self.referent_table = referent_table
        self.local_cols = local_cols
        self.remote_cols = remote_cols

    def execute(self, op: Operations):
        op.create_foreign_key(
            constraint_name=self.constraint_name,
            source_table=self.source_table,
            referent_table=self.referent_table,
            local_cols=self.local_cols,
            remote_cols=self.remote_cols
        )


class RenameSequenceMigrationAction(MigrationAction):
    def __init__(self, old_name: str, new_name: str):
        self.query = f'ALTER SEQUENCE "{old_name}" RENAME TO "{new_name}"'

    def execute(self, op: Operations):
        replaced = self.query.replace(":", "\\:")
        op.execute(replaced)


class TransferJSONDataMigrationAction(MigrationAction):
    def __init__(self, table: str, source: sa.Column, columns: List[Tuple[str, sa.Column]]):
        # Hack to transfer data if columns do not exist
        # Create new table with just required columns exist
        # SQLAlchemy does not allow the use table actions when those columns do not exist
        meta = sa.MetaData()
        temp_table = sa.Table(
            table,
            meta,
            source._copy(),
            *[column._copy() for _, column in columns]
        )
        self.query = temp_table.update().values(
            **{
                temp_table.columns[column.name].name: source[key].astext
                for key, column in columns
            }
        )

    def execute(self, op: Operations):
        op.execute(self.query)


class TransferColumnDataToJSONMigrationAction(MigrationAction):
    def __init__(self, table: str, source: sa.Column, columns: List[Tuple[str, sa.Column]]):
        # # Hack to transfer data if columns do not exist
        # # Create new table with just required columns exist
        # # SQLAlchemy does not allow the use table actions when those columns do not exist
        temp_table = sa.Table(
            table,
            sa.MetaData(),
            source._copy(),
            *[column._copy() for _, column in columns]
        )
        copied_source = temp_table.columns[source.name]
        results = []
        for key, value in columns:
            results.append(key)
            results.append(temp_table.columns[value.name])
        self.query = temp_table.update().values(
            **{
                copied_source.name: copied_source + sa.func.jsonb_build_object(
                    *results
                )
            }
        )

    def execute(self, op: Operations):
        op.execute(self.query)


class AddEmptyAttributeToJSONMigrationAction(MigrationAction):
    def __init__(self, table: str, source: sa.Column, key: str):
        temp_table = sa.Table(
            table,
            sa.MetaData(),
            source._copy(),
        )
        copied_source = temp_table.columns[source.name]
        self.query = temp_table.update().values(
            **{
                copied_source.name: copied_source + sa.func.jsonb_build_object(
                    key, None
                )
            }
        )

    def execute(self, op: Operations):
        op.execute(self.query)


class RenameJSONAttributeMigrationAction(MigrationAction):
    def __init__(self, table: str, source: sa.Column, old_key: str, new_key: str):
        temp_table = sa.Table(
            table,
            sa.MetaData(),
            source._copy(),
        )
        copied_source = temp_table.columns[source.name]
        self.query = temp_table.update().values(
            **{
                copied_source.name: copied_source - old_key + sa.func.jsonb_build_object(new_key,
                                                                                         copied_source[old_key])
            }
        ).where(copied_source.has_key(old_key))

    def execute(self, op: Operations):
        op.execute(self.query)


class RemoveJSONAttributeMigrationAction(MigrationAction):
    def __init__(self, table: str, source: sa.Column, key: str):
        temp_table = sa.Table(
            table,
            sa.MetaData(),
            source._copy(),
        )
        copied_source = temp_table.columns[source.name]
        self.query = temp_table.update().values(
            **{
                copied_source.name: copied_source - key
            }
        ).where(copied_source.has_key(key))

    def execute(self, op: Operations):
        op.execute(self.query)


class MigrationHandler:
    def __init__(self):
        self.migrations: List[MigrationAction] = []
        self.foreign_key_migration: List[MigrationAction] = []

    def add_action(self, action: MigrationAction, foreign_key: bool = False):
        if foreign_key:
            self.foreign_key_migration.append(action)
        else:
            self.migrations.append(action)

        return self

    def has_constraint_been_dropped(self, constraint_name: str):
        for migration in self.migrations:
            if isinstance(migration, DropConstraintMigrationAction):
                if migration.constraint_name == constraint_name:
                    return True
        for migration in self.foreign_key_migration:
            if isinstance(migration, DropConstraintMigrationAction):
                if migration.constraint_name == constraint_name:
                    return True
        return False

    def run_migrations(self, op: Operations):
        for migration in self.migrations:
            migration.execute(op)
        for migration in self.foreign_key_migration:
            migration.execute(op)



import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB

from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name
from spinta.components import Context
from spinta.types.text.components import Text


@commands.prepare.register(Context, PostgreSQL, Text)
def prepare(context: Context, backend: PostgreSQL, dtype: Text, **kwargs):
    prop = dtype.prop
    name = get_column_name(prop)

    return sa.Column(name, JSONB, unique=dtype.unique, nullable=False, default=sa.text("'{}'"), server_default=sa.text("'{}'"))



import uuid
from typing import Iterator
from typing import List
from typing import Optional

from spinta import commands
from spinta.backends import SelectTree, get_property_base_model
from spinta.backends import get_model_reserved_props
from spinta.backends.helpers import get_ns_reserved_props
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import get_select_tree
from spinta.backends.helpers import select_only_props
from spinta.components import Action, pagination_enabled, Node
from spinta.components import Context
from spinta.components import Model
from spinta.components import UrlParams
from spinta.formats.components import Format
from spinta.types.datatype import Array, ArrayBackRef, BackRef
from spinta.types.datatype import DataType
from spinta.types.datatype import ExternalRef
from spinta.types.datatype import File
from spinta.types.datatype import Inherit
from spinta.types.datatype import Object
from spinta.types.datatype import Ref
from spinta.types.text.components import Text
from spinta.ufuncs.basequerybuilder.ufuncs import Star
from spinta.utils.data import take
from spinta.utils.nestedstruct import get_separated_name


def _get_dtype_header(
    dtype: DataType,
    select: SelectTree,
    name: str,
    langs: List = None
) -> Iterator[str]:
    if isinstance(dtype, Object):
        for prop, sel in select_only_props(
            dtype.prop,
            take(dtype.properties).keys(),
            dtype.properties,
            select,
        ):
            name_ = get_separated_name(dtype, name, prop.name)
            yield from _get_dtype_header(prop.dtype, sel, name_, langs)

    elif isinstance(dtype, Array):
        name_ = name + '[]'
        yield from _get_dtype_header(dtype.items.dtype, select, name_, langs)

    elif isinstance(dtype, ArrayBackRef):
        name_ = name + '[]'
        yield from _get_dtype_header(dtype.refprop.dtype, select, name_, langs)

    elif isinstance(dtype, BackRef):
        yield from _get_dtype_header(dtype.refprop.dtype, select, name, langs)

    elif isinstance(dtype, File):
        yield get_separated_name(dtype, name, '_id')
        yield get_separated_name(dtype, name, '_content_type')

    elif isinstance(dtype, ExternalRef):
        if select is None or select == {'*': {}}:
            if dtype.model.given.pkeys or dtype.explicit:
                props = dtype.refprops
            else:
                props = [dtype.model.properties['_id']]
            processed_props = []
            for prop in props:
                processed_props.append(get_separated_name(dtype, name, prop.place))

            for key, prop in dtype.properties.items():
                for processed_name in _get_dtype_header(prop.dtype, select, get_separated_name(dtype, name, key), langs):
                    if processed_name not in processed_props:
                        processed_props.append(processed_name)
            yield from processed_props
        else:
            for prop, sel in select_only_props(
                dtype.prop,
                dtype.model.properties.keys(),
                dtype.model.properties,
                select,
            ):
                name_ = get_separated_name(dtype, name, prop.name)
                yield from _get_dtype_header(prop.dtype, sel, name_, langs)

    elif isinstance(dtype, Ref):
        if select is None or select == {'*': {}}:
            if not dtype.inherited:
                yield get_separated_name(dtype, name, '_id')
            for key, prop in dtype.properties.items():
                yield from _get_dtype_header(prop.dtype, select, get_separated_name(dtype, name, key), langs)
        else:
            for prop, sel in select_only_props(
                dtype.prop,
                dtype.model.properties.keys(),
                dtype.model.properties,
                select,
            ):
                name_ = get_separated_name(dtype, name, prop.name)
                yield from _get_dtype_header(prop.dtype, sel, name_, langs)
    elif isinstance(dtype, Text):
        if select is None or select == {'*': {}}:
            yield_text_count = 0
            if langs:
                if len(langs) == 1 and isinstance(langs[0], Star):
                    for lang in dtype.langs.keys():
                        yield_text_count += 1
                        yield get_separated_name(dtype, name, lang)
                elif len(langs) == 1:
                    yield_text_count += 1
                    yield name
                else:
                    for lang in langs:
                        if lang in dtype.langs:
                            yield_text_count += 1
                            yield get_separated_name(dtype, name, lang)
            else:
                yield_text_count += 1
                yield name

            if yield_text_count == 0:
                yield name
        else:
            for prop, sel in select_only_props(
                dtype.prop,
                dtype.langs.keys(),
                dtype.langs,
                select,
            ):
                name_ = get_separated_name(dtype, name, prop.name)
                yield from _get_dtype_header(prop.dtype, sel, name_, langs)
    elif isinstance(dtype, Inherit):
        if select and select != {'*': {}}:
            properties = {}
            for sel in select.keys():
                base_model = get_property_base_model(dtype.prop.model, sel)
                properties.update(base_model.properties)
            for prop, sel in select_only_props(
                dtype.prop,
                properties.keys(),
                properties,
                select,
            ):
                name_ = get_separated_name(dtype, name, prop.name)
                yield from _get_dtype_header(prop.dtype, sel, name_, langs)
        else:
            yield name
    else:
        yield name


def _get_model_header(
    model: Model,
    names: List[str],
    select: SelectTree,
    reserved: List[str],
    langs: List
) -> List[str]:
    if select is None or select == {'*': {}}:
        keys = reserved + names
    else:
        keys = names
    props = model.properties
    props_ = select_only_props(
        model,
        keys,
        props,
        select,
        reserved=True,
    )
    for prop, sel in props_:
        yield from _get_dtype_header(prop.dtype, sel, prop.name, langs)


def get_model_tabular_header(
    context: Context,
    model: Model,
    action: Action,
    params: UrlParams,
    *,
    reserved: Optional[List[str]] = None,
) -> List[str]:
    if reserved is None:
        if model.name == '_ns':
            reserved = get_ns_reserved_props(action)
        else:
            reserved = get_model_reserved_props(action, pagination_enabled(model, params))

    prop_select = params.select_props
    func_select = params.select_funcs
    select = get_select_tree(context, action, prop_select)

    if model.name == '_ns':
        names = get_select_prop_names(
            context,
            model,
            model.properties,
            action,
            select,
            auth=False,
        )
    else:
        names = get_select_prop_names(
            context,
            model,
            model.properties,
            action,
            select,
            reserved=reserved,
        )

    langs = params.lang
    if params.changes:
        langs = [Star()]

    header = list(_get_model_header(model, names, select, reserved, langs))

    if func_select is not None:
        for key in func_select.keys():
            header.append(key)

    return header


def rename_page_col(rows):
    for row in rows:
        yield {'_page.next' if k == '_page' else k: v for k, v in row.items()}


@commands.gen_object_id.register(Context, Format, Node)
def gen_object_id(context: Context, fmt: Format, node: Node):
    return str(uuid.uuid4())


@commands.gen_object_id.register(Context, Format)
def gen_object_id(context: Context, fmt: Format):
    return str(uuid.uuid4())



import json
import os
from typing import TypedDict, Dict, Union, List, Callable

from spinta.exceptions import FileNotFound
from spinta.utils.nestedstruct import get_root_attr


class MigrateTableRename(TypedDict):
    name: str
    new_name: str
    columns: Dict[str, str]


class MigrateRename:
    tables: Dict[str, MigrateTableRename]

    def __init__(self, rename_src: Union[str, dict]):
        self.tables = {}
        self.parse_rename_src(rename_src)

    def insert_table(self, table_name: str):
        self.tables[table_name] = MigrateTableRename(
            name=table_name,
            new_name=table_name,
            columns={}
        )

    def insert_column(self, table_name: str, column_name: str, new_column_name: str):
        if table_name not in self.tables.keys():
            self.insert_table(table_name)
        if column_name == "":
            self.tables[table_name]["new_name"] = new_column_name
        else:
            self.tables[table_name]["columns"][column_name] = new_column_name

    def get_column_name(self, table_name: str, column_name: str, root_only: bool = False, root_value: str = ""):
        # If table does not have renamed, return given column
        if table_name not in self.tables:
            return column_name

        table = self.tables[table_name]
        columns = table["columns"]

        if column_name in columns:
            return columns[column_name]

        # If column was not directly set, and it cannot be mapped through root node, return it
        if not root_only:
            return column_name

        root_attr = get_root_attr(column_name, initial_root=root_value)
        for old_column_name, new_column_name in columns.items():
            target_root_attr = get_root_attr(old_column_name, initial_root=root_value)
            if root_attr == target_root_attr:
                new_name = get_root_attr(new_column_name, initial_root=root_value)
                return new_name
        return column_name

    def get_old_column_name(self, table_name: str, column_name: str, root_only: bool = False, root_value: str = ""):
        if table_name not in self.tables:
            return column_name

        table = self.tables[table_name]
        given_name = get_root_attr(column_name, initial_root=root_value) if root_only else column_name
        for old_column_column, new_column_name in table["columns"].items():
            target_name = get_root_attr(new_column_name, initial_root=root_value) if root_only else new_column_name

            if target_name == given_name:
                old_name = get_root_attr(old_column_column, initial_root=root_value) if root_only else old_column_column
                return old_name
        return column_name

    def get_table_name(self, table_name: str):
        if table_name in self.tables.keys():
            return self.tables[table_name]["new_name"]
        return table_name

    def get_old_table_name(self, table_name: str):
        for key, data in self.tables.items():
            if data["new_name"] == table_name:
                return key
        return table_name

    def parse_rename_src(self, rename_src: Union[str, dict]):
        if rename_src:
            if isinstance(rename_src, str):
                if os.path.exists(rename_src):
                    with open(rename_src, 'r') as f:
                        data = json.loads(f.read())
                        for table, table_data in data.items():
                            self.insert_table(table)
                            for column, column_data in table_data.items():
                                self.insert_column(table, column, column_data)
                else:
                    raise FileNotFound(file=rename_src)
            else:
                for table, table_data in rename_src.items():
                    self.insert_table(table)
                    for column, column_data in table_data.items():
                        self.insert_column(table, column, column_data)


class MigrateMeta:
    plan: bool
    autocommit: bool
    rename: MigrateRename
    datasets: List[str]
    migration_extension: Callable

    def __init__(self, plan: bool, autocommit: bool, rename: MigrateRename, datasets: List[str] = None,
                 migration_extension: Callable = None):
        self.plan = plan
        self.rename = rename
        self.autocommit = autocommit
        self.datasets = datasets
        self.migration_extension = migration_extension



from spinta.formats.components import Format


class Xlsx(Format):
    content_type = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    accept_types = {
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    }
    params = {}



from spinta import commands
from spinta.backends.helpers import validate_and_return_begin
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.migrate import MigrateRename, MigrateMeta
from spinta.components import Context
from spinta.exceptions import ModelNotFound, PropertyNotFound
from spinta.manifests.components import Manifest


def _validate_migrate_rename(context: Context, rename: MigrateRename, manifest: Manifest):
    tables = rename.tables.values()
    for table in tables:
        models = commands.get_models(context, manifest)
        if table["new_name"] not in models.keys():
            raise ModelNotFound(model=table["new_name"])
        model = commands.get_model(context, manifest, table["new_name"])
        for column in table["columns"].values():
            if column not in model.flatprops.keys():
                raise PropertyNotFound(property=column)


@commands.migrate.register(Context, Manifest, MigrateMeta)
def migrate(context, manifest, migrate_meta):
    with context:
        require_auth(context)
        _validate_migrate_rename(context, migrate_meta.rename, manifest)
        backend = manifest.backend
        if backend:
            context.attach(f'transaction.{backend.name}', validate_and_return_begin, context, backend)
            commands.migrate(context, manifest, backend, migrate_meta)




from typing import Dict, Union
from spinta.components import UrlParams, Property, Model


def extract_params_sort_values(
    model: Model,
    params: UrlParams
) -> Union[type(None), Dict[str, Property]]:
    """
    A hack to iterate through sort expression without using resolvers

    return sort properties
    if key starts with '-' it means that it's descending

    if returns None, then it could not parse something
    """
    if not params.sort:
        return {}

    result = {}
    for sort_ast in params.sort:
        negative = False
        if sort_ast['name'] == 'negative':
            negative = True
            sort_ast = sort_ast['args'][0]
        elif sort_ast['name'] == 'positive':
            sort_ast = sort_ast['args'][0]
        for sort in sort_ast['args']:
            if isinstance(sort, str):
                if sort in model.properties.keys():
                    by = sort if not negative else f'-{sort}'
                    prop = model.properties[sort]
                    result[by] = prop
            else:
                return None

    return result



import datetime
from typing import List, Any

import sqlalchemy as sa
import tqdm
from typer import echo

from spinta import commands
from spinta.auth import authorized
from spinta.backends.constants import BackendFeatures
from spinta.cli.helpers.errors import ErrorCounter
from spinta.cli.helpers.push import prepare_data_for_push_state
from spinta.cli.helpers.push.utils import extract_state_page_id_key, construct_where_condition_from_page
from spinta.commands.read import PaginationMetaData, get_paginated_values
from spinta.components import Context, Action, Config
from spinta.components import Model
from spinta.components import Page, get_page_size, Property, PageBy
from spinta.utils.response import get_request


def _build_push_state_sync_url(
    server: str,
    model: str,
    page: str,
    page_columns: List[str],
    limit: int
):
    base = f'{server}/{model}/:format/json?'

    required = [
        '_id',
        '_revision',
        '_page',
        'checksum()'
    ]

    base = f'{base}select({",".join(required)}'
    for page_column in page_columns:
        base = f'{base},{page_column}'
    base = f'{base})&sort(_id)'

    if page:
        base = f'{base}&page(\'{page}\')'

    if limit:
        base = f'{base}&limit({limit})'

    return base


def _fetch_all_model_data(
    config: Config,
    model: Model,
    client,
    server: str,
    error_counter: ErrorCounter,
    timeout: tuple[float, float],
    initial_page_data: Any = None
):
    limit = config.sync_page_size
    page_hash = ''
    page_columns = []
    page = commands.create_page(model.page, initial_page_data)
    if model.backend.supports(BackendFeatures.PAGINATION) and page.by and page.enabled:
        for page_by in page.by.values():
            page_columns.append(page_by.prop.name)
    while True:
        url = _build_push_state_sync_url(
            server=server,
            model=model.model_type(),
            page=page_hash,
            page_columns=page_columns,
            limit=limit
        )
        status_code, resp = get_request(
            client,
            url,
            error_counter=error_counter,
            timeout=timeout
        )
        if status_code == 200:
            data = resp['_data']
            if not data:
                break

            if '_page' in resp:
                page_hash = resp['_page']['next']

            for row in data:
                yield row

        else:
            break


def _get_state_rows_with_id(
    context: Context,
    table: sa.Table,
    size: int
) -> sa.engine.LegacyCursorResult:
    conn = context.get('push.state.conn')

    model_page = Page()
    model_page.size = size + 1
    prop = Property()
    prop.name = 'id'
    model_page.by = {
        'id': PageBy(prop)
    }
    order_by = sa.sql.expression.nullslast(sa.asc(table.c['id']))

    page_meta = PaginationMetaData(
        page_size=size,
    )

    while not page_meta.is_finished:
        page_meta.is_finished = True
        from_cond = construct_where_condition_from_page(model_page, table, prefix='')

        if from_cond is not None:
            stmt = sa.select([table]).where(
                from_cond
            ).order_by(order_by).limit(model_page.size)
            rows = conn.execute(
                stmt
            )
        else:
            stmt = sa.select([table]).order_by(order_by).limit(model_page.size)
            rows = conn.execute(
                stmt
            )

        yield from get_paginated_values(model_page, page_meta, rows, extract_state_page_id_key)


def _delete_row_from_push_state(
    conn: sa.engine.Connection,
    table: sa.Table,
    id_: str
):
    conn.execute(
        table.delete().
        where(table.c.id == id_)
    )


def _update_row_from_push_state(
    context: Context,
    model: Model,
    conn: sa.engine.Connection,
    table: sa.Table,
    id_: str,
    target_row: dict,
    state_row: dict
):
    target_checksum = target_row.get('checksum()')
    state_checksum = getattr(state_row, 'checksum')

    # Check if target db and pushed data is the same

    page_keys = []
    reserved_keys = [
        '_id',
        '_revision',
        'checksum()'
    ]

    page_mapping = {}
    for key in target_row.keys():
        if key not in reserved_keys:
            page_keys.append(key)

    skip_update = False
    if target_checksum == state_checksum:
        skip_update = True
        # Check if pushed state did not have errors while pushing already existing data
        if getattr(state_row, 'error') or getattr(state_row, 'data'):
            skip_update = False

        # Check if revisions match, they need match in order to do proper updates
        if getattr(state_row, 'revision') != target_row['_revision']:
            skip_update = False

        # Check if page key values match (could be cases when migrating from old to new versions, or changing page keys)
        if skip_update:
            for key in page_keys:
                if target_row[key] != getattr(state_row, f'page.{key}'):
                    skip_update = False
                    break

    # Skip update if nothing has changed
    if skip_update:
        return

    if page_keys:
        page_mapping = {key: target_row[key] for key in page_keys}
        page_mapping = prepare_data_for_push_state(context, model, page_mapping)
        page_mapping = {f'page.{key}': value for key, value in page_mapping.items()}

    conn.execute(
        table.update().
        where(table.c.id == id_).values(
            {
                'revision': target_row.get('_revision'),
                'checksum': target_checksum,
                'pushed': datetime.datetime.now(),
                'error': False,
                'data': None,
                **page_mapping
            }
        )
    )


def _insert_row_to_push_state(
    context: Context,
    model: Model,
    conn: sa.engine.Connection,
    table: sa.Table,
    target_row: dict
):
    page_keys = []
    reserved_keys = [
        '_id',
        '_revision',
        'checksum()'
    ]

    page_mapping = {}
    for key in target_row.keys():
        if key not in reserved_keys:
            page_keys.append(key)

    if page_keys:
        page_mapping = {key: target_row[key] for key in page_keys}
        page_mapping = prepare_data_for_push_state(context, model, page_mapping)
        page_mapping = {f'page.{key}': value for key, value in page_mapping.items()}

    conn.execute(
        table.insert().
        values(
            {
                'id': target_row['_id'],
                'revision': target_row.get('_revision'),
                'checksum': target_row.get('checksum()'),
                'pushed': datetime.datetime.now(),
                'error': False,
                'data': None,
                **page_mapping
            }
        )
    )


def sync_push_state(
    context: Context,
    client,
    server: str,
    models: List[Model],
    error_counter: ErrorCounter,
    no_progress_bar: bool,
    metadata: sa.MetaData,
    timeout: tuple[float, float]
):
    config = context.get('config')
    conn = context.get('push.state.conn')
    counters = {}
    if not no_progress_bar:
        counters = {
            '_total': tqdm.tqdm(desc='SYNCHRONIZING PUSH STATE', ascii=True)
        }
        echo()

    for model in models:
        model_name = model.model_type()
        primary_keys = model.external.pkeys
        size = get_page_size(config, model)
        skip_model = False
        # Check permissions
        for key in primary_keys:
            is_authorized = authorized(context, key, action=Action.SEARCH)
            if not is_authorized:
                echo(f"SKIPPED PUSH STATE '{model.model_type()}' MODEL SYNC, NO PERMISSION.")
                skip_model = True
                break

        if skip_model:
            continue

        if not no_progress_bar:
            counters[model_name] = tqdm.tqdm(desc=model_name, ascii=True)

        model_table = metadata.tables[model.name]
        target_data = _fetch_all_model_data(
            config=config,
            model=model,
            client=client,
            server=server,
            error_counter=error_counter,
            timeout=timeout
        )
        state_data = _get_state_rows_with_id(
            context=context,
            table=model_table,
            size=size
        )
        target_row = next(target_data, None)
        state_row = next(state_data, None)
        while target_row is not None or state_row is not None:
            target_id = target_row['_id'] if target_row is not None else None
            state_id = state_row['id'] if state_row is not None else None

            if not no_progress_bar:
                counters['_total'].update(1)
                counters[model_name].update(1)

            if target_row is None:
                _delete_row_from_push_state(conn, model_table, state_id)
                state_row = next(state_data, None)
                continue

            if state_row is None:
                _insert_row_to_push_state(context, model, conn, model_table, target_row)
                target_row = next(target_data, None)
                continue

            if target_id == state_id:
                _update_row_from_push_state(context, model, conn, model_table, state_id, target_row, state_row)
                target_row = next(target_data, None)
                state_row = next(state_data, None)
            elif target_id < state_id:
                _insert_row_to_push_state(context, model, conn, model_table, target_row)
                target_row = next(target_data, None)
            else:
                _delete_row_from_push_state(conn, model_table, state_id)
                state_row = next(state_data, None)

        if model_name in counters:
            counters[model_name].close()

    if '_total' in counters:
        counters['_total'].close()



from typing import List

from spinta import commands
from spinta.backends.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name as gcn
from spinta.components import Property, Model
from spinta.exceptions import PropertyNotFound
from spinta.types.datatype import ArrayBackRef, DataType, Ref, BackRef, String, ExternalRef, Array
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB

from spinta.types.text.components import Text


def convert_str_to_columns(table: sa.Table, prop: Property, columns: List[str]) -> List[sa.Column]:
    return_list = []
    for column in columns:
        return_list.append(convert_str_to_column(table, prop, column))
    return return_list


def convert_str_to_column(table: sa.Table, prop: Property, column: str) -> sa.Column:
    if column in table.c:
        return table.c[column]
    else:
        raise PropertyNotFound(prop.dtype)


def get_table(backend: PostgreSQL, prop: Property) -> sa.Table:
    if prop.list:
        table = backend.get_table(prop.list, TableType.LIST)
    else:
        table = backend.get_table(prop)
    return table


@commands.get_column.register(PostgreSQL, Property)
def get_column(backend: PostgreSQL, prop: Property, **kwargs):
    return commands.get_column(backend, prop.dtype, **kwargs)


@commands.get_column.register(PostgreSQL, DataType)
def get_column(backend: PostgreSQL, dtype: DataType, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    if table is None:
        table = get_table(backend, prop)
    column_name = gcn(dtype.prop)
    column = convert_str_to_column(table, prop, column_name)
    if isinstance(column.type, (sa.types.JSON, JSONB)) and dtype.prop.place != dtype.prop.name:
        return column[dtype.prop.name].label(dtype.prop.place)
    return column


@commands.get_column.register(PostgreSQL, Array)
def get_column(backend: PostgreSQL, dtype: Array, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    if table is None:
        table = get_table(backend, prop)
    column_name = gcn(dtype.prop)
    column = convert_str_to_column(table, prop, column_name)
    return column


@commands.get_column.register(PostgreSQL, ArrayBackRef)
def get_column(backend: PostgreSQL, dtype: ArrayBackRef, table: sa.Table = None, **kwargs):
    return commands.get_column(backend, dtype.items.dtype, **kwargs)


@commands.get_column.register(PostgreSQL, String)
def get_column(backend: PostgreSQL, dtype: String, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    parent = prop.parent

    if table is None:
        table = get_table(backend, prop)
    column_name = gcn(parent if parent and isinstance(parent.dtype, Text) else dtype.prop)
    column = convert_str_to_column(table, prop, column_name)

    if parent and isinstance(parent.dtype, Text):
        return column[dtype.prop.name].label(dtype.prop.place.replace('@', '.'))

    if isinstance(column.type, (sa.types.JSON, JSONB)) and dtype.prop.place != dtype.prop.name:
        return column[dtype.prop.name].label(dtype.prop.place)
    return column


@commands.get_column.register(PostgreSQL, Ref)
def get_column(backend: PostgreSQL, dtype: Ref, table: sa.Table = None, **kwargs):
    column = gcn(dtype.prop, replace=dtype.inherited)
    prop = dtype.prop
    if table is None:
        table = get_table(backend, prop)
    if prop.list is not None:
        column = '_id'
    else:
        column += '._id'
    return convert_str_to_column(table, prop, column)


@commands.get_column.register(PostgreSQL, ExternalRef)
def get_column(backend: PostgreSQL, dtype: Ref, table: sa.Table = None, **kwargs):
    column_prefix = gcn(dtype.prop, replace=dtype.inherited)
    prop = dtype.prop
    if table is None:
        table = get_table(backend, prop)

    result = []
    for ref_prop in dtype.refprops:
        ref_prop_ = ref_prop
        table_ = table
        column = ref_prop.name
        if prop.list is None:
            column = f'{column_prefix}.{ref_prop.name}'
        if ref_prop.name in dtype.properties:
            if isinstance(ref_prop.dtype, type(dtype.properties[ref_prop.name].dtype)):
                ref_prop_ = dtype.properties[ref_prop.name]
                table_ = get_table(backend, ref_prop_)
                column = ref_prop_.place
        result.append(convert_str_to_column(table_, ref_prop_, column))
    return result


@commands.get_column.register(PostgreSQL, BackRef)
def get_column(backend: PostgreSQL, dtype: BackRef, table: sa.Table = None, **kwargs):
    prop = dtype.prop

    if prop.list is not None:
        return None

    r_prop = dtype.refprop
    columns = []
    if table is None:
        table = get_table(backend, dtype.prop)
    if commands.identifiable(r_prop):
        column = '_id'
        return convert_str_to_column(table, prop, column)
    else:
        for refprop in r_prop.dtype.refprops:
            columns.append(refprop.name)
        return convert_str_to_columns(table, prop, columns)


@commands.get_column.register(PostgreSQL, Property, Model)
def get_column(backend: PostgreSQL, prop: Property, model: Model, table: sa.Table = None, **kwargs):
    return commands.get_column(backend, prop.dtype, model, table=table, **kwargs)


@commands.get_column.register(PostgreSQL, DataType, Model)
def get_column(backend: PostgreSQL, prop: Property, model: Model, table: sa.Table = None, **kwargs):
    return None


@commands.get_column.register(PostgreSQL, Ref, Model)
def get_column(backend: PostgreSQL, dtype: Ref, model: Model, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    if table is None:
        table = backend.get_table(model)
    column = '_id'
    return convert_str_to_column(table, prop, column)


@commands.get_column.register(PostgreSQL, ExternalRef, Model)
def get_column(backend: PostgreSQL, dtype: ExternalRef, model: Model, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    if table is None:
        table = backend.get_table(model)

    columns = []
    for refprop in prop.dtype.refprops:
        columns.append(refprop.name)
    return convert_str_to_columns(table, prop, columns)






import logging

from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.yaml.components import InlineManifest
from spinta.manifests.yaml.components import YamlManifest
from spinta.manifests.yaml.helpers import read_inline_manifest_schemas
from spinta.manifests.yaml.helpers import read_manifest_schemas
from spinta.manifests.yaml.helpers import read_freezed_manifest_schemas

log = logging.getLogger(__name__)


@commands.load.register(Context, YamlManifest)
def load(
    context: Context,
    manifest: YamlManifest,
    *,
    into: Manifest = None,
    freezed: bool = False,
    rename_duplicates: bool = False,
    load_internal: bool = True,
    full_load: bool = False
):
    if load_internal:
        target = into or manifest
        if not commands.has_model(context, target, '_schema'):
            store = context.get('store')
            commands.load(context, store.internal, into=target, full_load=full_load)

    if freezed:
        if into:
            log.info(
                'Loading freezed manifest %r into %r from %s.',
                manifest.name,
                into.name,
                manifest.path.resolve(),
            )
        else:
            log.info(
                'Loading freezed manifest %r from %s.',
                manifest.name,
                manifest.path.resolve(),
            )
        schemas = read_freezed_manifest_schemas(manifest)
    else:
        if into:
            log.info(
                'Loading manifest %r into %r from %s.',
                manifest.name,
                into.name,
                manifest.path.resolve(),
            )
        else:
            log.info(
                'Loading manifest %r from %s.',
                manifest.name,
                manifest.path.resolve(),
            )
        schemas = read_manifest_schemas(manifest)

    if into:
        load_manifest_nodes(context, into, schemas, source=manifest)
    else:
        load_manifest_nodes(context, manifest, schemas)


@commands.load.register(Context, InlineManifest)
def load(
    context: Context,
    manifest: InlineManifest,
    *,
    into: Manifest = None,
    freezed: bool = True,
    rename_duplicates: bool = False,
    load_internal: bool = True,
    full_load: bool = False
):
    assert freezed, (
        "InlineManifest does not have unfreezed version of manifest."
    )

    if load_internal:
        target = into or manifest
        if not commands.has_model(context, target, '_schema'):
            store = context.get('store')
            commands.load(context, store.internal, into=target, full_load=full_load)

    if into:
        log.info(
            'Loading freezed manifest %r into %r from %s.',
            manifest.name,
            into.name,
            '<inline>',
        )
        schemas = read_inline_manifest_schemas(manifest)
        load_manifest_nodes(context, into, schemas, source=manifest)
    else:
        log.info(
            'Loading freezed manifest %r from %s.',
            manifest.name,
            '<inline>',
        )
        schemas = read_inline_manifest_schemas(manifest)
        load_manifest_nodes(context, manifest, schemas)

    for source in manifest.sync:
        commands.load(
            context, source,
            into=into or manifest,
            freezed=freezed,
            rename_duplicates=rename_duplicates,
            load_internal=load_internal,
            full_load=full_load
        )



from typing import Any

from dask.dataframe import DataFrame

from spinta.components import Model, Property
from spinta.core.ufuncs import Env, Expr
from spinta.datasets.backends.dataframe.components import DaskBackend
from spinta.exceptions import UnknownMethod
from spinta.ufuncs.basequerybuilder.components import Selected
from spinta.utils.schema import NA


class DaskDataFrameQueryBuilder(Env):
    backend: DaskBackend
    model: Model
    dataframe: DataFrame

    def init(self, backend: DaskBackend, dataframe: DataFrame):
        return self(
            backend=backend,
            dataframe=dataframe,
            resolved={},
            selected=None,
            sort={
                "desc": [],
                "asc": []
            },
            limit=None,
            offset=None,
        )

    def build(self, where):
        if self.selected is None:
            self.call('select', Expr('select'))
        df = self.dataframe

        if self.limit is not None:
            df = df.head(self.limit, npartitions=-1, compute=False)

        if self.offset is not None:
            df = df.loc[self.offset:]

        if where is not None:
            df = df[where]
        return df

    def execute(self, expr: Any):
        expr = self.call('_resolve_unresolved', expr)
        return super().execute(expr)

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownMethod(name=expr.name, expr=str(expr(*args, **kwargs)))


class DaskSelected(Selected):
    # Item name in select list.
    item: str = None
    # Model property if a property is selected.
    prop: Property = None
    # A value or an Expr for further processing on selected value.
    prep: Any = NA



import uuid
from pathlib import Path

import click

from spinta import commands
from spinta.auth import auth_server_keys_exists, client_name_exists, get_clients_path, ensure_client_folders_exist
from spinta.auth import create_client_file
from spinta.auth import gen_auth_server_keys
from spinta.backends.helpers import validate_and_return_transaction, validate_and_return_begin
from spinta.cli.helpers.upgrade.clients import requires_client_migration
from spinta.components import Config
from spinta.components import Context
from spinta.components import Store
from spinta.core.config import DEFAULT_CONFIG_PATH
from spinta.exceptions import ClientsMigrationRequired
from spinta.manifests.components import Manifest


def _ensure_config_dir(
    config: Config,
    path: Path,
    default_auth_client: str,
    *,
    verbose: bool = True,
):
    """Create default config_path directory if it does not exits"""

    # If a non default config directory is given, it should exist.
    if not path.exists() and path != DEFAULT_CONFIG_PATH:
        raise Exception(f"Config dir {path} does not exist!")


    clients_path = get_clients_path(config)
    # Check if client migrations are needed
    if requires_client_migration(clients_path):
        raise ClientsMigrationRequired()

    # Ensure all files/folders exist for clients operations
    ensure_client_folders_exist(clients_path)

    # Ensure auth server keys.
    if not auth_server_keys_exists(path):
        if verbose:
            click.echo(f"Initializing auth server keys: {path}")
        gen_auth_server_keys(path)

    # Ensure default client.
    if default_auth_client and not client_name_exists(clients_path, default_auth_client):
        if verbose:
            click.echo(f"Initializing default auth client: {path}")
        create_client_file(
            clients_path,
            default_auth_client,
            str(uuid.uuid4()),
            secret=None,
            scopes=[
                'spinta_getall',
                'spinta_getone',
                'spinta_search',
                'spinta_changes',
            ],
        )


def load_config(
    context: Context,
    *,
    verbose: bool = True,
    ensure_config_dir: bool = False,
) -> Config:
    config = context.get('config')
    commands.load(context, config)
    if ensure_config_dir:
        _ensure_config_dir(
            config,
            config.config_path,
            config.default_auth_client,
            verbose=verbose,
        )
    commands.check(context, config)
    return config


def load_store(
    context: Context,
    *,
    verbose: bool = True,
    ensure_config_dir: bool = False,
) -> Store:
    load_config(
        context,
        verbose=verbose,
        ensure_config_dir=ensure_config_dir,
    )
    store = context.get('store')
    commands.load(context, store)
    return store


def load_manifest(
    context: Context,
    *,
    store: Store = None,
    verbose: bool = True,
    ensure_config_dir: bool = False,
    rename_duplicates: bool = False,
    load_internal: bool = True,
    full_load: bool = False
) -> Store:
    if store is None:
        store = load_store(
            context,
            verbose=verbose,
            ensure_config_dir=ensure_config_dir,
        )
    if verbose:
        if store.manifest.path:
            click.echo(
                f"Loading {type(store.manifest).__name__} "
                f"manifest {store.manifest.name} "
                f"({store.manifest.path})..."
            )
        else:
            click.echo(
                f"Loading {type(store.manifest).__name__} "
                f"manifest {store.manifest.name}"
            )
    commands.load(
        context, store.manifest,
        rename_duplicates=rename_duplicates,
        load_internal=load_internal,
        full_load=full_load
    )
    commands.link(context, store.manifest)
    commands.check(context, store.manifest)
    return store


def prepare_manifest(
    context: Context,
    *,
    verbose: bool = True,
    ensure_config_dir: bool = False,
    full_load: bool = False
) -> Store:
    store = load_manifest(
        context,
        verbose=verbose,
        ensure_config_dir=ensure_config_dir,
        full_load=full_load
    )
    commands.wait(context, store)
    commands.prepare(context, store.manifest)
    return store


def attach_backends(context: Context, store: Store, manifest: Manifest) -> None:
    context.attach('transaction', validate_and_return_transaction, context, manifest.backend)
    backends = set()
    for backend in store.backends.values():
        backends.add(backend.name)
        context.attach(f'transaction.{backend.name}', validate_and_return_begin, context, backend)
    for backend in manifest.backends.values():
        backends.add(backend.name)
        context.attach(f'transaction.{backend.name}', validate_and_return_begin, context, backend)
    for dataset_ in commands.get_datasets(context, manifest).values():
        for resource in dataset_.resources.values():
            if resource.backend and resource.backend.name not in backends:
                backends.add(resource.backend.name)
                context.attach(f'transaction.{resource.backend.name}', validate_and_return_begin, context, resource.backend)


def attach_keymaps(context: Context, store: Store) -> None:
    for keymap in store.keymaps.values():
        context.attach(f'keymap.{keymap.name}', lambda: keymap)



from spinta.components import PageInfo, Page
from multipledispatch import dispatch


@dispatch(PageInfo)
def page_contains_unsupported_keys(page: PageInfo):
    allowed_types = get_allowed_page_property_types()
    for prop in page.keys.values():
        if not isinstance(prop.dtype, allowed_types):
            return True
    return False


@dispatch(Page)
def page_contains_unsupported_keys(page: Page):
    allowed_types = get_allowed_page_property_types()
    for page_by in page.by.values():
        if not isinstance(page_by.prop.dtype, allowed_types):
            return True
    return False


def get_allowed_page_property_types():
    from spinta.types.datatype import Integer, Number, String, Date, Time, DateTime, PrimaryKey, UUID
    return Integer, Number, String, Date, DateTime, Time, PrimaryKey, UUID



import csv
import datetime
import json
import pathlib
from functools import lru_cache
from typing import List, AsyncIterator

from multipledispatch import dispatch

from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.cli.helpers.export.backends.postgresql.components import PostgresqlExportMetadata, TableData
from spinta.cli.helpers.export.components import CounterManager
from spinta.components import Context, Model, Action, DataItem
from spinta.types.datatype import Array, DataType, File
from spinta.utils.aiotools import anext, achain
from spinta.utils.itertools import ensure_list
from spinta.utils.json import fix_data_for_json


def generate_export_metadata(
        model: Model,
        name: str,
        file: object,
        fieldnames: list,
        counter: CounterManager
) -> PostgresqlExportMetadata:
    counter.add_counter(
        model,
        name
    )

    writer = csv.DictWriter(file, fieldnames=fieldnames)
    return PostgresqlExportMetadata(
        name=name,
        model=model,
        writer=writer,
    )


def generate_table_data(dtype: DataType, data: object):
    return TableData(
        name=generate_file_name(dtype),
        prop=dtype.prop,
        data=data
    )


@dispatch(Array)
@lru_cache
def generate_file_name(dtype: Array) -> str:
    return f'{dtype.prop.model.model_type()}.{dtype.prop.place}.array'


@dispatch(File)
@lru_cache
def generate_file_name(dtype: File) -> str:
    return f'{dtype.prop.model.model_type()}.{dtype.prop.place}.file'


@dispatch(DataType)
def generate_file_name(dtype: DataType):
    return None


@dispatch(Array)
@lru_cache
def generate_file_name(dtype: Array):
    return f'{dtype.prop.model.model_type()}.{dtype.prop.place}.array'


@dispatch(File)
@lru_cache
def generate_file_name(dtype: File):
    return f'{dtype.prop.model.model_type()}.{dtype.prop.place}.file'


@dispatch(Context, File, PostgreSQL)
def generate_csv_headers(context: Context, dtype: DataType, backend: PostgreSQL):
    raise Exception("NOT IMPLEMENTED")


@dispatch(Context, File, PostgreSQL)
def generate_csv_headers(context: Context, dtype: File, backend: PostgreSQL):
    raise Exception("NOT IMPLEMENTED")


@dispatch(Context, Array, PostgreSQL)
def generate_csv_headers(context: Context, dtype: Array, backend: PostgreSQL):
    columns = commands.prepare(context, backend, dtype.items)
    columns = ensure_list(columns)

    return [
        '_txn',
        '_rid',
        *[column.name for column in columns]
    ]


def split_data(data: dict) -> (dict, List[TableData]):
    other_dict = {k: v for k, v in data.items() if not isinstance(v, TableData)}
    table_list = [v for v in data.values() if isinstance(v, TableData)]
    return other_dict, table_list


def prepare_changelog(data: dict, _id: int, txn: str, created: datetime.datetime) -> dict:
    reserved_data = {k: v for k, v in data.items() if k.startswith('_')}
    given_data = {k: v for k, v in data.items() if not k.startswith('_')}
    return {
        '_id': _id,
        '_revision': reserved_data['_revision'],
        '_txn': txn,
        '_rid': reserved_data['_id'],
        'datetime': created,
        'action': Action.INSERT.value,
        'data': json.dumps(fix_data_for_json(given_data))
    }


def generate_file_paths(model: Model, root_path: pathlib.Path) -> (dict, dict):
    directory = root_path / model.ns.name
    directory.mkdir(exist_ok=True, parents=True)

    default_paths = [
        model.model_type(),
        f'{model.model_type()}.changes'
    ]

    additional_paths = {
        prop: generate_file_name(prop.dtype) for prop in model.flatprops.values()
    }
    additional_paths = {
        path: prop for prop, path in additional_paths.items() if path is not None
    }

    extended_paths = default_paths + list(additional_paths.keys())
    return {
        path: f'{root_path / pathlib.Path(path)}.csv' for path in extended_paths
    }, additional_paths


async def extract_headers(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    dstream: AsyncIterator[DataItem],
    **kwargs
) -> (list, AsyncIterator[DataItem]):
    peek = await anext(dstream)
    data = commands.before_export(
        context,
        model,
        backend,
        data=peek,
        **kwargs
    )
    data_headers = split_data(data)[0].keys()
    dstream = achain([peek], dstream)
    return data_headers, dstream



import csv
from typing import Any
from multipledispatch import dispatch

from spinta.cli.helpers.export.components import CounterManager
from spinta.components import Model, Property
from spinta.types.datatype import DataType, Array


class TableData:
    name: str
    prop: Property
    data: Any

    def __init__(self, name: str, prop: Property, data: Any):
        self.name = name
        self.prop = prop
        self.data = data


class PostgresqlExportMetadata:
    name: str
    model: Model
    writer: csv.DictWriter

    def __init__(self, name: str, model: Model, writer: csv.DictWriter):
        self.name = name
        self.model = model
        self.writer = writer

    async def writeheader(self):
        await self.writer.writeheader()

    async def writerow(self, data: dict, counter: CounterManager):
        await self.writer.writerow(data)
        counter.update_specific(self.model, self.name, 1)

    async def write_table_data(self, table_data: TableData, data: dict, counter: CounterManager):
        counter.update_specific(self.model, self.name, 1)
        if not table_data.data or not isinstance(table_data.data, (list, dict)):
            return

        extra_fields = extract_extra_fields(table_data.prop.dtype, data)
        if isinstance(table_data.data, list):
            for value in table_data.data:
                await self.writer.writerow({
                    **extra_fields,
                    **value
                })
        else:
            await self.writer.writerow({
                **extra_fields,
                **table_data.data
            })


@dispatch(DataType)
def extract_extra_fields(dtype: DataType, data: dict) -> dict:
    return {}


@dispatch(Array)
def extract_extra_fields(dtype: Array, data: dict) -> dict:
    return {
        '_txn': data['_txn'],
        '_rid': data['_id']
    }



from typing import Union

from spinta.components import Property, FuncProperty
from spinta.core.ufuncs import ufunc, Expr, Bind
from spinta.types.datatype import Integer, String
from spinta.ufuncs.requestparamsbuilder.components import RequestParamsBuilder
from spinta.ufuncs.requestparamsbuilder.helpers import disable_params_pagination


@ufunc.resolver(RequestParamsBuilder, Expr)
def select(env: RequestParamsBuilder, expr: Expr):
    for arg in expr.args:
        result_name = str(arg)
        result = env.resolve(arg)
        env.call('select', result_name, result, given=arg)


@ufunc.resolver(RequestParamsBuilder, str, FuncProperty)
def select(env: RequestParamsBuilder, select_name: str, func_prop: FuncProperty, given: object = None, **kwargs):
    params = env.params

    if func_prop.func is None:
        func_prop.func = given

    if params.select_funcs is None:
        params.select_funcs = {}
    params.select_funcs[select_name] = func_prop

    if params.select_props is None:
        params.select_props = {}


@ufunc.resolver(RequestParamsBuilder, str, (Bind, Expr))
def select(env: RequestParamsBuilder, select_name: str, other: Union[Bind, Expr], **kwargs):
    params = env.params
    if params.select_props is None:
        params.select_props = {}
    params.select_props[select_name] = other

    if params.select_funcs is None:
        params.select_funcs = {}


@ufunc.resolver(RequestParamsBuilder)
def count(env: RequestParamsBuilder):
    disable_params_pagination(env.params)

    prop = Property()
    prop.name = 'count()'
    prop.place = 'count()'
    prop.title = ''
    prop.description = ''
    prop.model = env.params.model
    prop.dtype = Integer()
    prop.dtype.type = 'integer'
    prop.dtype.type_args = []
    prop.dtype.name = 'integer'
    prop.dtype.prop = prop

    return FuncProperty(func=None, prop=prop)


@ufunc.resolver(RequestParamsBuilder, Expr)
def checksum(env: RequestParamsBuilder, expr: Expr):
    prop = Property()
    prop.name = 'checksum()'
    prop.place = 'checksum()'
    prop.title = ''
    prop.description = ''
    prop.model = env.params.model
    prop.dtype = String()
    prop.dtype.type = 'string'
    prop.dtype.type_args = []
    prop.dtype.name = 'string'
    prop.dtype.prop = prop

    return FuncProperty(func=None, prop=prop)



import enum


class Access(enum.IntEnum):
    # Private properties can be accesses only if an explicit property scope is
    # given. I client does not have required scope, then private properties
    # can't bet selected, but can be used in query conditions, in sorting.
    private = 0

    # Property is exposed only to authorized user, who has access to model.
    # Authorization token is given manually to each user.
    protected = 1

    # Property can be accessed by anyone, but only after accepting terms and
    # conditions, that means, authorization is still needed and data can only be
    # used as specified in provided terms and conditions. Authorization token
    # can be obtained via WebUI.
    public = 2

    # Open data, anyone can access this data, no authorization is required.
    open = 3


class Level(enum.IntEnum):
    # Data do not exist.
    absent = 0

    # Data exists in any form, for example txt, pdf, etc...
    available = 1

    # Data is structured, for example xls, xml, etc...
    structured = 2

    # Data provided using an open format, csv, tsv, sql, etc...
    open = 3

    # Individual data objects have unique identifiers.
    identifiable = 4

    # Data is linked with a known vocabulary.
    linked = 5



from typing import Any

from spinta.components import Property
from spinta.ufuncs.resultbuilder.components import ResultBuilder
from sqlalchemy.engine.row import RowProxy


class SqlResultBuilder(ResultBuilder):
    this: Any           # A selected value.
    prop: Property      # Model property.
    data: RowProxy      # Whole row from database.

    def init(self, this: Any, prop: Property, data: RowProxy):
        return self(
            this=this,
            prop=prop,
            data=data,
        )



from multipledispatch import dispatch
from typing import Dict, Union, Any, Callable, Tuple
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set

import itertools

from spinta import commands
from spinta.components import Model, Property
from spinta.types.datatype import DataType, Ref, Partial, Object
from spinta.types.text.components import Text


SEP_GETTER_TYPE = Callable[[str], Tuple[Callable, str]]


@dispatch(type(None))
def _get_seperator(target):
    return None


@dispatch((Model, DataType))
def _get_seperator(target):
    return '.'


@dispatch(Property)
def _get_seperator(target: Property):
    return _get_seperator(target.dtype)


@dispatch(Text)
def _get_seperator(target: Text):
    return '@'


@dispatch((Model, Object, Partial), str)
def _get_child(parent: Union[Model, Object, Partial], target: str):
    if target in parent.properties:
        return parent.properties[target]


@dispatch(Property, str)
def _get_child(parent: Property, target: str):
    return _get_child(parent.dtype, target)


@dispatch(DataType, str)
def _get_child(parent: DataType, target: str):
    return None


@dispatch(type(None), str)
def _get_child(parent, target):
    return None


@dispatch(Ref, str)
def _get_child(parent: Ref, name: str):
    for refprop in parent.refprops:
        if refprop.name == name:
            return refprop

    if name in parent.properties:
        return parent.properties[name]

    if commands.identifiable(parent.prop) and name == '_id':
        return parent.model.properties['_id']

    if parent.model.external and parent.model.external.unknown_primary_key:
        return parent.model.properties['_id']


@dispatch(Text, str)
def _get_child(parent: Text, name: str):
    if name in parent.langs:
        return parent.langs[name]


def sepgetter(parent: Any = None, default_seperator='.') -> SEP_GETTER_TYPE:
    def _sepgetter(target: str):
        child = _get_child(parent, target)
        seperator = _get_seperator(parent)
        if seperator is None:
            seperator = default_seperator
        return sepgetter(child, default_seperator=default_seperator), seperator

    return _sepgetter


@dispatch(DataType, str, str)
def get_separated_name(parent: DataType, parent_name: str, child_name: str):
    seperator = _get_seperator(parent)
    return get_separated_name(seperator, parent_name, child_name)


@dispatch(str, str, str)
def get_separated_name(seperator: str, parent_name: str, child_name: str):
    if parent_name:
        return f'{parent_name}{seperator}{child_name}'
    return child_name


def flatten(value, sep_getter: SEP_GETTER_TYPE = sepgetter(), omit_none: bool = True):
    value, lists = _flatten(value, sep_getter, omit_none=omit_none)

    if value is None:
        for k, vals in lists:
            for v in vals:
                if v is not None or not omit_none:
                    yield from flatten(v, sep_getter, omit_none=omit_none)

    elif lists:
        keys, lists = zip(*lists)
        for vals in itertools.product(*lists):
            val = {
                k: v
                for k, v in zip(keys, vals) if v is not None or not omit_none
            }
            val.update(value)
            yield from flatten(val, sep_getter, omit_none=omit_none)

    else:
        yield value


def _flatten(value, sep_getter: SEP_GETTER_TYPE, key: str = '', omit_none: bool = True):
    if isinstance(value, dict):
        data = {}
        lists = []
        for k, v in value.items():
            if v is not None or not omit_none:
                new_sep_getter, seperator = sep_getter(k)
                new_name = get_separated_name(seperator, key, k)
                v, more = _flatten(v, new_sep_getter, new_name, omit_none=omit_none)
                data.update(v or {})
                lists += more
        return data, lists

    elif isinstance(value, (list, Iterator)):
        if value:
            if key:
                key = f'{key}[]'
            return None, [(key, value)] if value is not None or not omit_none else []
        else:
            return None, []

    else:
        return {key: value} if value is not None or not omit_none else {}, []


def build_select_tree(select: List[str]) -> Dict[str, Set[Optional[str]]]:
    tree: Dict[str, Set[Optional[str]]] = {}
    for name in select:
        split = name, None
        while len(split) == 2:
            name, node = split
            if name not in tree:
                tree[name] = set()
            if node:
                tree[name].add(node)
            split = name.rsplit('.', 1)
    return tree


def flat_dicts_to_nested(value, list_keys: list = None):
    if list_keys is None:
        list_keys = []
    res = {}

    def recursive_nesting(data, res_, keys: list, depth: int):
        if depth >= len(keys):
            return

        key = keys[depth]
        place = '.'.join(keys[:depth + 1])

        is_array = place in list_keys
        is_last = len(keys) - 1 == depth

        if is_array:
            if key not in res_:
                res_[key] = []

            data_ = data
            if not isinstance(data_, list):
                data_ = [data_]
            for item in data_:
                new_dict = {}
                recursive_nesting(item, new_dict, keys, depth + 1)
                res_[key].append(new_dict)
        else:
            if key not in res_:
                res_[key] = {}
            recursive_nesting(data, res_[key], keys, depth + 1)

        if is_last:
            if key in res_ and res_[key] and isinstance(res_[key], dict):
                res_[key].update(data)
            else:
                res_[key] = data

    for k, v in dict(value).items():
        names = k.split('.')
        recursive_nesting(v, res, names, 0)
    return res


def flatten_value(value, parent: Property, sep=".", key=""):
    value, _ = _flatten(value, sepgetter(parent, default_seperator=sep), key)
    return value


def get_root_attr(value: str, initial_root: str = "") -> str:
    value = value.split('@')[0]
    if not initial_root:
        return value.split('.')[0]

    contains_prefix = False
    value = value.replace(initial_root, '', 1)

    if value.startswith('.'):
        contains_prefix = True
        value = value.removeprefix('.')

    value = value.split('.')[0]
    root = f'{initial_root}.' if contains_prefix else initial_root

    return root + value


def get_last_attr(value: str):
    return value.split("@")[-1].split(".")[-1]


def extract_list_property_names(
    model: Model,
    properties: List[str],
) -> List[str]:
    list_keys = []
    for key in properties:
        if key in model.flatprops:
            prop = model.flatprops[key]
            if prop.list is not None and prop.list.place not in list_keys:
                list_keys.append(prop.list.place)
    return list_keys



import re

from spinta import commands
from spinta.components import Context
from spinta.components import Config
from spinta.core.enums import Level
from spinta.datasets.components import Dataset, Resource
from spinta.exceptions import InvalidName

namespace_re = re.compile(r'^[a-z](_?[a-z0-9]+)*(/[a-z](_?[a-z0-9]+)+)*$')


def check_dataset_name(name: str):
    return namespace_re.match(name) is not None


@commands.check.register()
def check(context: Context, dataset: Dataset):
    for resource in dataset.resources.values():
        commands.check(context, resource)

    config: Config = context.get('config')

    if config.check_names:
        name = dataset.name
        if namespace_re.match(name) is None:
            raise InvalidName(dataset, name=name, type='namespace')


@commands.check.register()
def check(context: Context, resource: Resource):
    pass


@commands.identifiable.register()
def identifiable(dataset: Dataset):
    return dataset.level is None or dataset.level >= Level.identifiable


@commands.identifiable.register()
def identifiable(resource: Resource):
    return resource.level is None or resource.level >= Level.identifiable



from spinta import commands
from spinta.components import Context, MetaData
from spinta.dimensions.param.helpers import finalize_param_link
from spinta.manifests.components import Manifest, get_manifest_object_names


@commands.link.register(Context, Manifest)
def link(context: Context, manifest: Manifest):
    for node in get_manifest_object_names():
        for obj in commands.get_nodes(context, manifest, node).values():
            commands.link(context, obj)

    finalize_param_link(context, manifest)


@commands.link.register(Context, MetaData)
def link(context: Context, node: MetaData):
    pass



from copy import deepcopy
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List

import requests
import sqlalchemy as sa
import tqdm

from spinta import commands
from spinta.cli.helpers.data import ModelRow, count_rows, read_model_data, filter_allowed_props_for_model, \
    filter_dict_by_keys
from spinta.cli.helpers.errors import ErrorCounter
from spinta.cli.helpers.push.components import State, PushRow, PUSH_NOW
from spinta.cli.helpers.push.delete import get_deleted_rows
from spinta.cli.helpers.push.error import get_rows_with_errors, get_rows_with_errors_counts
from spinta.cli.helpers.push.utils import get_data_checksum, extract_state_page_keys, update_model_page_with_new, \
    construct_where_condition_from_page
import spinta.cli.push as cli_push
from spinta.commands.read import get_page, PaginationMetaData, get_paginated_values
from spinta.components import Context, pagination_enabled
from spinta.components import Model
from spinta.components import Page, get_page_size
from spinta.core.ufuncs import Expr
from spinta.ufuncs.basequerybuilder.components import QueryParams
from spinta.utils.itertools import peek


def _iter_model_rows(
    context: Context,
    models: List[Model],
    counts: Dict[str, int],
    metadata: sa.MetaData,
    limit: int = None,
    *,
    initial_page_data: dict = None,
    stop_on_error: bool = False,
    no_progress_bar: bool = False,
    push_counter: tqdm.tqdm = None,
) -> Iterator[ModelRow]:
    if initial_page_data is None:
        initial_page_data = {}

    params = QueryParams()
    params.push = True
    for model in models:

        model_push_counter = None
        if not no_progress_bar:
            count = counts.get(model.name)
            model_push_counter = tqdm.tqdm(desc=model.name, ascii=True, total=count, leave=False)

        if pagination_enabled(model):
            page = commands.create_page(model.page, initial_page_data.get(model.model_type(), None))
            rows = _read_rows_by_pages(
                context,
                model,
                page,
                metadata,
                limit,
                stop_on_error,
                push_counter,
                model_push_counter,
                params=params
            )
            for row in rows:
                yield row
        else:
            stream = read_model_data(
                context,
                model,
                limit,
                stop_on_error,
                params=params
            )
            for item in stream:
                if push_counter:
                    push_counter.update(1)
                if model_push_counter:
                    model_push_counter.update(1)
                yield PushRow(model, item)

        if model_push_counter is not None:
            model_push_counter.close()


def _get_model_rows(
    context: Context,
    models: List[Model],
    metadata: sa.MetaData,
    limit: int = None,
    *,
    initial_page_data: dict,
    stop_on_error: bool = False,
    no_progress_bar: bool = False,
    error_counter: ErrorCounter = None,
) -> Iterator[PushRow]:
    counts = (
        count_rows(
            context,
            models,
            limit,
            stop_on_error=stop_on_error,
            error_counter=error_counter,
            initial_page_data=initial_page_data
        )
        if not no_progress_bar
        else {}
    )
    push_counter = None
    if not no_progress_bar:
        push_counter = tqdm.tqdm(desc='PUSH', ascii=True, total=sum(counts.values()))
    rows = _iter_model_rows(
        context,
        models,
        counts,
        metadata,
        limit,
        initial_page_data=initial_page_data,
        stop_on_error=stop_on_error,
        no_progress_bar=no_progress_bar,
        push_counter=push_counter,
    )
    for row in rows:
        yield row

    if push_counter is not None:
        push_counter.close()


def read_rows(
    context: Context,
    client: requests.Session,
    server: str,
    models: List[Model],
    state: State,
    limit: int = None,
    *,
    timeout: tuple[float, float],
    stop_on_error: bool = False,
    retry_count: int = 5,
    no_progress_bar: bool = False,
    error_counter: ErrorCounter = None,
    initial_page_data: dict = None
) -> Iterator[PushRow]:
    if initial_page_data is None:
        initial_page_data = {}

    yield from _get_model_rows(
        context,
        models,
        state.metadata,
        limit,
        stop_on_error=stop_on_error,
        no_progress_bar=no_progress_bar,
        error_counter=error_counter,
        initial_page_data=initial_page_data
    )

    yield PUSH_NOW
    yield from get_deleted_rows(
        models,
        context,
        state.metadata,
        no_progress_bar=no_progress_bar,
    )

    for i in range(1, retry_count + 1):
        yield PUSH_NOW
        counts = (
            get_rows_with_errors_counts(
                models,
                context,
                state.metadata
            )
        )
        total_count = sum(counts.values())

        if total_count > 0:
            yield from get_rows_with_errors(
                client,
                server,
                models,
                context,
                state.metadata,
                counts,
                retry=i,
                timeout=timeout,
                no_progress_bar=no_progress_bar,
                error_counter=error_counter,
            )
        else:
            break


def _read_model_data_with_page(
    context: Context,
    model: Model,
    model_page: Page,
    limit: int = None,
    stop_on_error: bool = False,
    params: QueryParams = None
) -> Iterable[Dict[str, Any]]:

    if limit is None:
        query = None
    else:
        query = Expr('limit', limit)

    stream = get_page(
        context,
        model,
        model.backend,
        model_page,
        query,
        limit,
        params=params
    )

    if stop_on_error:
        stream = peek(stream)
    else:
        try:
            stream = peek(stream)
        except Exception:
            cli_push.log.exception(f"Error when reading data from model {model.name}")
            return

    prop_filer, needs_filtering = filter_allowed_props_for_model(model)
    for item in stream:
        if needs_filtering:
            item = filter_dict_by_keys(prop_filer, item)
        yield item


def _read_rows_by_pages(
    context: Context,
    model: Model,
    page: Page,
    metadata: sa.MetaData,
    limit: int = None,
    stop_on_error: bool = False,
    push_counter: tqdm.tqdm = None,
    model_push_counter: tqdm.tqdm = None,
    params: QueryParams = None
) -> Iterator[PushRow]:
    conn = context.get('push.state.conn')
    config = context.get('config')

    size = get_page_size(config, model)
    model_table = metadata.tables[model.name]
    state_rows = _get_state_rows_with_page(
        context,
        deepcopy(page),
        model_table,
        size
    )
    rows = _read_model_data_with_page(
        context,
        model,
        deepcopy(page),
        limit,
        stop_on_error,
        params
    )
    total_count = 0
    data_push_count = 0
    state_push_count = 0
    data_row = next(rows, None)
    state_row = next(state_rows, None)

    while True:
        if limit:
            if total_count >= limit:
                break
            total_count += 1

        update_counter = True

        if data_push_count >= size or state_push_count >= size:
            state_rows = _get_state_rows_with_page(
                context,
                deepcopy(page),
                model_table,
                size
            )
            rows = _read_model_data_with_page(
                context,
                model,
                deepcopy(page),
                limit,
                stop_on_error
            )

            data_push_count = 0
            state_push_count = 0
            data_row = next(rows, None)
            state_row = next(state_rows, None)

        if data_row is None and state_row is None:
            break

        if data_row is not None and state_row is not None:
            row = PushRow(model, data_row)
            row.op = 'insert'
            row.checksum = get_data_checksum(row.data, model)

            equals = _compare_data_with_state_rows(data_row, state_row, model_table)
            if equals:
                row.op = 'patch'
                row.saved = True
                row.data['_revision'] = state_row[model_table.c.revision]
                if state_row[model_table.c.checksum] != row.checksum or not _compare_data_with_state_row_keys(data_row, state_row, model_table, page):
                    yield row

                data_push_count += 1
                state_push_count += 1
                update_model_page_with_new(page, model_table, data_row=data_row)
                data_row = next(rows, None)
                state_row = next(state_rows, None)

            else:
                delete_cond = _compare_for_delete_row(state_row, data_row, model_table, page)
                if delete_cond:
                    conn.execute(
                        sa.update(model_table).where(model_table.c.id == state_row["id"]).values(pushed=None)
                    )

                    state_push_count += 1
                    update_model_page_with_new(page, model_table, state_row=state_row)
                    state_row = next(state_rows, None)
                    update_counter = False
                else:
                    yield row

                    data_push_count += 1
                    update_model_page_with_new(page, model_table, data_row=data_row)
                    data_row = next(rows, None)
        else:
            if data_row is not None:
                row = PushRow(model, data_row)
                row.op = 'insert'
                row.checksum = get_data_checksum(row.data, model)
                yield row

                data_push_count += 1
                update_model_page_with_new(page, model_table, data_row=data_row)
                data_row = next(rows, None)
            elif state_row is not None:
                conn.execute(
                    sa.update(model_table).where(model_table.c.id == state_row["id"]).values(pushed=None)
                )

                state_push_count += 1
                update_model_page_with_new(page, model_table, state_row=state_row)
                state_row = next(state_rows, None)
                update_counter = False

        if update_counter:
            if push_counter is not None:
                push_counter.update(1)
            if model_push_counter is not None:
                model_push_counter.update(1)


def _get_state_rows_with_page(
    context: Context,
    model_page: Page,
    table: sa.Table,
    size: int
) -> sa.engine.LegacyCursorResult:
    conn = context.get('push.state.conn')
    model_page.size = size + 1
    order_by = []

    for by, page_by in model_page.by.items():
        if by.startswith('-'):
            order_by.append(sa.sql.expression.nullsfirst(sa.desc(table.c[f"page.{page_by.prop.name}"])))
        else:
            order_by.append(sa.sql.expression.nullslast(sa.asc(table.c[f"page.{page_by.prop.name}"])))

    page_meta = PaginationMetaData(
        page_size=size,
    )

    while not page_meta.is_finished:
        page_meta.is_finished = True
        from_cond = construct_where_condition_from_page(model_page, table)

        if from_cond is not None:
            stmt = sa.select([table]).where(
                from_cond
            ).order_by(*order_by).limit(model_page.size)
            rows = conn.execute(
                stmt
            )
        else:
            stmt = sa.select([table]).order_by(*order_by).limit(model_page.size)
            rows = conn.execute(
                stmt
            )

        yield from get_paginated_values(model_page, page_meta, rows, extract_state_page_keys)


def _compare_data_with_state_rows(
    data_row: Any,
    state_row: Any,
    model_table: sa.Table,
):
    return data_row["_id"] == state_row[model_table.c["id"]]


def _compare_data_with_state_row_keys(
    data_row: Any,
    state_row: Any,
    model_table: sa.Table,
    page: Page
):
    equals = True
    decoded = data_row.get("_page")
    for i, (by, page_by) in enumerate(page.by.items()):
        state_val = state_row[model_table.c[f"page.{page_by.prop.name}"]]
        data_val = decoded[i]
        if state_val != data_val:
            equals = False
            break
    return equals


def _compare_for_delete_row(
    state_row: Any,
    data_row: Any,
    model_table: sa.Table,
    page: Page,
):
    delete_cond = False

    if '_page' in data_row:
        decoded = data_row['_page']
        for i, (by, page_by) in enumerate(page.by.items()):
            state_val = state_row[model_table.c[f"page.{page_by.prop.name}"]]
            data_val = decoded[i]
            compare = True
            if state_val is not None and data_val is not None:
                if by.startswith('-'):
                    compare = state_val > data_val
                else:
                    compare = state_val < data_val
            if state_val != data_val:
                delete_cond = compare
                break
    else:
        delete_cond = True
    return delete_cond



from spinta import commands
from spinta.backends import Backend
from spinta.components import Context
from spinta.ufuncs.resultbuilder.components import ResultBuilder


@commands.get_result_builder.register(Context, Backend)
def get_result_builder(
    context: Context,
    backend: Backend
):
    return ResultBuilder(context)



from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Array


@commands.check.register(Context, Array)
def check(context: Context, dtype: Array):
    if dtype.items is not None:
        commands.check(context, dtype.items)



from collections.abc import Generator
from typing import Optional, Any

import datetime
import uuid
import hashlib
import decimal
from uuid import UUID

import msgpack
import sqlalchemy as sa

from spinta import commands
from spinta.cli.helpers.data import ensure_data_dir
from spinta.components import Config
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.datasets.keymaps.components import KeyMap
from sqlalchemy.dialects.sqlite import insert

from spinta.datasets.keymaps.sync import KeymapData
from spinta.exceptions import KeyMapGivenKeyMissmatch

from multipledispatch import dispatch


class SqlAlchemyKeyMap(KeyMap):
    dsn: str = None

    sync_table_name: str = '_synchronize'
    sync_transaction_size: int = None

    def __init__(self, dsn: str = None):
        self.dsn = dsn
        self.engine = None
        self.metadata = None
        self.conn = None

    def __enter__(self):
        assert self.dsn is not None
        assert self.conn is None
        self.conn = self.engine.connect()
        return self

    def __exit__(self, *exc):
        self.conn.close()
        self.conn = None

    def get_table(self, name):
        if name not in self.metadata.tables:
            if name == self.sync_table_name:
                table = sa.Table(
                    name, self.metadata,
                    sa.Column('model', sa.Text, primary_key=True),
                    sa.Column('cid', sa.BIGINT),
                    sa.Column('updated', sa.DateTime),
                )
            else:
                table = sa.Table(
                    name, self.metadata,
                    sa.Column('key', sa.Text, primary_key=True),
                    sa.Column('hash', sa.Text, unique=True, index=True),
                    sa.Column('value', sa.LargeBinary),
                )
            table.create(checkfirst=True)
        return self.metadata.tables[name]

    def encode(self, name: str, value: object, primary_key=None) -> Optional[str]:
        # Make value msgpack serializable.

        hash_return = _hash_value(value)
        if hash_return is None:
            return None

        value, hashed = hash_return

        table = self.get_table(name)
        current_key = self.conn.execute(
            sa.select([table.c.key]).where(
                table.c.hash == hashed
            )
        ).scalar()

        # Key was found in the table and no primary was given
        if current_key is not None and primary_key is None:
            return current_key

        if primary_key is not None:
            # Given primary key matches found key
            if primary_key == current_key:
                return primary_key

            if current_key is not None and current_key != primary_key:
                raise KeyMapGivenKeyMissmatch(
                    name=name,
                    given_key=primary_key,
                    found_key=current_key
                )

            current_key = primary_key

        if current_key is None:
            current_key = str(uuid.uuid4())

        self.conn.execute(table.insert(), {
            'key': current_key,
            'hash': hashed,
            'value': value,
        })
        return current_key

    def decode(self, name: str, key: str) -> object:
        table = self.get_table(name)
        query = sa.select([table.c.value]).where(table.c.key == key)
        value = self.conn.execute(query).scalar()
        value = msgpack.loads(value, raw=False)
        return value

    def contains(self, name: str, value: Any) -> bool:
        result = _hash_value(value)

        if result is None:
            return False

        encoded_value, encoded_hash = result

        table = self.get_table(name)
        query = sa.select([sa.func.count()]).where(
            sa.and_(
                table.c.value == encoded_value,
                table.c.hash == encoded_hash
            )
        )
        return self.conn.execute(query).scalar() > 0

    def get_last_synced_id(self, name: str) -> object:
        table = self.get_table(self.sync_table_name)
        query = sa.select([table.c.cid]).where(table.c.model == name)
        value = self.conn.execute(query).scalar()
        return value

    def update_sync_data(self, name: str, cid: Any, time: datetime.datetime):
        table = self.get_table(self.sync_table_name)
        query = insert(table).values(model=name, cid=cid, updated=time)
        query = query.on_conflict_do_update(
            index_elements=[table.c.model],
            set_=dict(cid=cid, updated=time)
        )
        self.conn.execute(query)

    def synchronize(self, name: str, value: Any, primary_key: str):
        table = self.get_table(name)
        hash_return = _hash_value(value)
        if hash_return is None:
            return None
        else:
            value, hashed = hash_return

        where_condition = sa.or_(table.c.key == primary_key, table.c.hash == hashed)
        select_query = sa.select([sa.func.count()]).select_from(table).where(where_condition)
        count = self.conn.execute(select_query).scalar()
        should_insert = True
        if count == 1:
            should_insert = False
            update_query = sa.update(table).values(key=primary_key, hash=hashed, value=value).where(where_condition)
            self.conn.execute(update_query)
        else:
            delete_query = sa.delete(table).where(where_condition)
            self.conn.execute(delete_query)

        if should_insert:
            query = insert(table).values(key=primary_key, hash=hashed, value=value)
            self.conn.execute(query)

    def has_synced_before(self) -> bool:
        table = self.get_table(self.sync_table_name)
        count = self.conn.execute(sa.func.count(table.c.model)).scalar()
        return count != 0


def _hash_value(value):
    if value is None:
        return None

    if isinstance(value, (list, tuple)):
        value = [_encode_value(k) for k in value if k is not None]
        if len(value) == 0:
            return None
    else:
        value = _encode_value(value)
    value = msgpack.dumps(value, strict_types=True)
    return value, hashlib.sha1(value).hexdigest()


@dispatch((datetime.datetime, datetime.date, datetime.time))
def _encode_value(value: (datetime.datetime, datetime.date, datetime.time)):
    return value.isoformat()


@dispatch(decimal.Decimal)
def _encode_value(value: decimal.Decimal):
    return float(value)


@dispatch(tuple)
def _encode_value(value: tuple):
    return list(value)


@dispatch(UUID)
def _encode_value(value: UUID):
    return str(value)


@dispatch(object)
def _encode_value(value: object):
    return value


@commands.configure.register(Context, SqlAlchemyKeyMap)
def configure(context: Context, keymap: SqlAlchemyKeyMap):
    rc: RawConfig = context.get('rc')
    config: Config = context.get('config')
    dsn = rc.get('keymaps', keymap.name, 'dsn', required=True)
    sync_transaction_size = rc.get('keymaps', keymap.name, 'sync_transaction_size', default=10000, cast=int)
    ensure_data_dir(config.data_path)
    dsn = dsn.format(data_dir=config.data_path)
    if dsn.startswith('sqlite:///'):
        dsn = dsn.replace('sqlite:///', 'sqlite+spinta:///')
    keymap.sync_transaction_size = sync_transaction_size
    keymap.dsn = dsn


@commands.prepare.register(Context, SqlAlchemyKeyMap)
def prepare(context: Context, keymap: SqlAlchemyKeyMap, **kwargs):
    keymap.engine = sa.create_engine(keymap.dsn)
    keymap.metadata = sa.MetaData(keymap.engine)


@commands.sync.register(Context, SqlAlchemyKeyMap)
def sync(context: Context, keymap: SqlAlchemyKeyMap, *, data: Generator[KeymapData]):
    transaction_size = keymap.sync_transaction_size
    transaction = None
    try:
        transaction = keymap.conn.begin()
        for i, row in enumerate(data):
            if transaction_size is not None and i % transaction_size == 0 and i != 0:
                if transaction.is_active:
                    transaction.commit()

                transaction = keymap.conn.begin()

            keymap.synchronize(row.key, row.value, row.identifier)
            yield row
    finally:
        if transaction is not None and transaction.is_active:
            transaction.commit()



from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.dict.components import DictManifest
from spinta.manifests.dict.helpers import read_schema


@commands.load.register(Context, DictManifest)
def load(
    context: Context,
    manifest: DictManifest,
    *,
    into: Manifest = None,
    freezed: bool = True,
    rename_duplicates: bool = False,
    load_internal: bool = True,
    full_load=False
):
    if load_internal:
        target = into or manifest
        if not commands.has_model(context, target, '_schema'):
            store = context.get('store')
            commands.load(context, store.internal, into=target, full_load=full_load)

    if manifest.path is None:
        return

    dataset_name = context.get("rc").get("given_dataset_name")
    schemas = read_schema(manifest.format, manifest.path, dataset_name)

    if into:
        load_manifest_nodes(context, into, schemas, source=manifest)
    else:
        load_manifest_nodes(context, manifest, schemas)

    for source in manifest.sync:
        commands.load(
            context, source,
            into=into or manifest,
            freezed=freezed,
            rename_duplicates=rename_duplicates,
            load_internal=load_internal,
            full_load=full_load
        )



import contextlib
from typing import Any
from typing import Dict
from typing import Optional
from typing import Set

from spinta.backends.constants import BackendOrigin, BackendFeatures


class Backend:
    metadata = {
        'name': 'backend',
    }

    type: str
    name: str
    origin: BackendOrigin
    features: Set[BackendFeatures] = set()

    # Original configuration values given in manifest, this is used to restore
    # manifest back to its original form.
    config: Dict[str, Any]

    available: bool = True

    def __repr__(self):
        return (
            f'<{self.__class__.__module__}.{self.__class__.__name__}'
            f'(name={self.name!r}) at 0x{id(self):02x}>'
        )

    @contextlib.contextmanager
    def transaction(self):
        raise NotImplementedError

    @contextlib.contextmanager
    def begin(self):
        raise NotImplementedError

    def bootstrapped(self):
        raise NotImplementedError

    # Checks if backend supports specific feature
    def supports(self, feature: BackendFeatures) -> bool:
        return feature in self.features


SelectTree = Optional[Dict[str, 'SelectTree']]



from spinta import commands
from spinta.auth import authorized
from spinta.components import Context, Action, Namespace
from spinta.manifests.components import Manifest


@commands.authorize.register(Context, Action, Namespace, Manifest)
def authorize(context: Context, action: Action, ns: Namespace, manifest: Manifest):
    authorized(context, ns, action, throw=True)



from spinta import commands
from spinta.backends import Backend
from spinta.cli.helpers.errors import cli_error
from spinta.components import Context, Model, DataSubItem, Property
from spinta.exceptions import NotImplementedFeature
from spinta.formats.components import Format
from spinta.types.datatype import DataType, Ref, ExternalRef, Denorm, Object
from spinta.utils.data import take
from spinta.utils.nestedstruct import flatten_value


@commands.export_data.register(Context, Model, Format)
def export_data(context: Context, model: Model, fmt: Format, *, data: object, **kwargs):
    cli_error(
        f"Export for {type(fmt)!r} Format is supported yet."
    )


@commands.export_data.register(Context, Model, Backend)
def export_data(context: Context, model: Model, backend: Backend, *, data: object, **kwargs):
    cli_error(
        f"Export for {backend.type!r} Backend is supported yet."
    )


@commands.before_export.register(Context, Model, Backend)
def before_export(
    context: Context,
    model: Model,
    backend: Backend,
    *,
    data: DataSubItem
):
    patch = take(['_id'], data.patch)
    patch['_revision'] = take('_revision', data.patch, data.saved)
    for prop in take(model.properties).values():
        if not prop.dtype.inherited:
            prop_data = data[prop.name]
            value = commands.before_export(
                context,
                prop.dtype,
                backend,
                data=prop_data,
            )
            patch.update(value)
    return patch


@commands.before_export.register(Context, Property, Backend)
def before_export(
    context: Context,
    prop: Property,
    backend: Backend,
    *,
    data: DataSubItem
):
    return before_export(context, prop.dtype, backend, data=data)


@commands.before_export.register(Context, DataType, Backend)
def before_export(
    context: Context,
    dtype: DataType,
    backend: Backend,
    *,
    data: DataSubItem
):
    return take(all, {dtype.prop.place: data.patch})


@commands.before_export.register(Context, Ref, Backend)
def before_export(
    context: Context,
    dtype: Ref,
    backend: Backend,
    *,
    data: DataSubItem,
) -> dict:
    # If patch is None, it means that parent was set to null, meaning all children should also be set to null

    if data.patch is None:
        patch = {}

        if not dtype.inherited:
            patch[f'{dtype.prop.place}._id'] = None

        for prop in dtype.properties.values():
            value = commands.before_export(
                context,
                prop.dtype,
                backend,
                data=data,
            )
            patch.update(value)
        return patch

    patch = flatten_value(data.patch, dtype.prop)
    return {
        f'{dtype.prop.place}.{k}': v for k, v in patch.items() if k
    }


@commands.before_export.register(Context, ExternalRef, Backend)
def before_export(
    context: Context,
    dtype: ExternalRef,
    backend: Backend,
    *,
    data: DataSubItem,
) -> dict:
    # If patch is None, it means that parent was set to null, meaning all children should also be set to null
    if data.patch is None:
        patch = {}
        if not dtype.inherited:
            if dtype.explicit or not dtype.model.external.unknown_primary_key:
                for ref_prop in dtype.refprops:
                    patch[f'{dtype.prop.place}.{ref_prop.name}'] = None
            else:
                patch[f'{dtype.prop.place}._id'] = None

        for prop in dtype.properties.values():
            value = commands.before_export(
                context,
                prop.dtype,
                backend,
                data=data,
            )
            patch.update(value)

        return patch

    patch = flatten_value(data.patch, dtype.prop)
    return {
        f'{dtype.prop.place}.{k}': v for k, v in patch.items() if k
    }


@commands.before_export.register(Context, Denorm, Backend)
def before_export(
    context: Context,
    dtype: Denorm,
    backend: Backend,
    *,
    data: DataSubItem,
) -> dict:
    # If patch is None, it means that all children should be set to null
    if data.patch is None:
        patch = commands.before_export(context, dtype.rel_prop.dtype, backend, data=data)
        return {
            key.replace(dtype.rel_prop.place, dtype.prop.place): value
            for key, value in patch.items() if key != dtype.rel_prop.place
        }

    patch = flatten_value(data.patch, dtype.prop)
    key = dtype.prop.place.split('.', maxsplit=1)[-1]
    if patch.get(key):
        return {dtype.prop.place: patch.get(key)}
    else:
        return {}


@commands.before_export.register(Context, Object, Backend)
def before_export(
    context: Context,
    dtype: Object,
    backend: Backend,
    *,
    data: DataSubItem,
) -> dict:
    patch = {}
    # If patch is None, then just set every child to None
    if data.patch is None:
        for prop in dtype.properties.values():
            value = commands.before_export(
                context,
                prop.dtype,
                backend,
                data=data,
            )
            patch.update(value)
        return patch

    for prop in dtype.properties.values():
        value = commands.before_export(
            context,
            prop.dtype,
            backend,
            data=data[prop.name],
        )
        patch.update(value)
    return patch


@commands.validate_export_output.register(Context, Backend, type(None))
def validate_export_output(context: Context, backend: Backend, output):
    cli_error(
        "Output argument is required (`--output`)."
    )


@commands.validate_export_output.register(Context, Format, type(None))
def validate_export_output(context: Context, backend: Format, output):
    cli_error(
        "Output argument is required (`--output`)."
    )


@commands.validate_export_output.register(Context, Backend, object)
def validate_export_output(context: Context, backend: Backend, output):
    return


@commands.validate_export_output.register(Context, Format, object)
def validate_export_output(context: Context, fmt: Format, output):
    return



from __future__ import annotations

from typing import Any
from typing import List
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import overload

import sqlalchemy as sa
from sqlalchemy.sql.functions import Function

from spinta.auth import authorized
from spinta.components import Action, Page
from spinta.components import Property
from spinta.core.ufuncs import Bind, GetAttr
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import Negative
from spinta.core.ufuncs import Unresolved
from spinta.core.ufuncs import ufunc
from spinta.datasets.backends.sql.helpers import dialect_specific_desc, dialect_specific_asc, \
    contains_geometry_flip_function, dialect_specific_geometry_flip
from spinta.datasets.backends.sql.ufuncs.query.components import SqlQueryBuilder
from spinta.dimensions.enum.helpers import prepare_enum_value
from spinta.exceptions import PropertyNotFound, SourceCannotBeList
from spinta.types.datatype import DataType, Denorm, Object
from spinta.types.datatype import PrimaryKey
from spinta.types.datatype import Ref
from spinta.types.datatype import String
from spinta.types.datatype import UUID
from spinta.types.geometry.components import Geometry
from spinta.types.text.components import Text
from spinta.types.text.helpers import determine_language_property_for_text
from spinta.ufuncs.basequerybuilder.components import LiteralProperty, Selected, Flip
from spinta.ufuncs.basequerybuilder.helpers import get_language_column, process_literal_value
from spinta.ufuncs.basequerybuilder.ufuncs import Star
from spinta.ufuncs.components import ForeignProperty
from spinta.utils.data import take
from spinta.utils.itertools import flatten
from spinta.utils.schema import NA


def _gather_selected_properties(env: SqlQueryBuilder):
    result = []
    if env.selected:
        for selected in env.selected.values():
            if selected and selected.prop:
                result.append(selected.prop)
    return result


@ufunc.resolver(SqlQueryBuilder, str, object, names=['eq', 'ne'])
def eq(env: SqlQueryBuilder, op: str, field: str, value: Any):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    return env.call(op, Bind(field), value)


def _sa_compare(op: str, column: sa.Column, value: Any):
    if op == 'eq':
        return column == value

    if op == 'ne':
        return column != value

    if op == 'lt':
        return column < value

    if op == 'le':
        return column <= value

    if op == 'gt':
        return column > value

    if op == 'ge':
        return column >= value

    if op == 'contains':
        return column.contains(value)

    if op == 'startswith':
        return column.startswith(value)

    raise NotImplementedError


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]


T = TypeVar('T')


def _prepare_value(prop: Property, value: T) -> Union[T, List[T]]:
    return prepare_enum_value(prop, value)


@ufunc.resolver(SqlQueryBuilder, Bind, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, field: Bind, value: Any):
    prop = env.model.properties[field.name]
    value = _prepare_value(prop, value)
    return env.call(op, prop.dtype, value)


@ufunc.resolver(SqlQueryBuilder, GetAttr, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, attr: GetAttr, value: Any):
    resolved = env.call('_resolve_getattr', attr)
    return env.call(op, resolved, value)


@ufunc.resolver(SqlQueryBuilder, Bind, list, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, field: Bind, value: List[Any]):
    prop = env.model.properties[field.name]
    value = list(flatten(_prepare_value(prop, v) for v in value))
    return env.call(op, prop.dtype, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, fpr: ForeignProperty, value: Any):
    value = _prepare_value(fpr.right.prop, value)
    return env.call(op, fpr, fpr.right, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, list, names=COMPARE)
def compare(
    env: SqlQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    value: List[Any],
):
    value = list(flatten(_prepare_value(fpr.right.prop, v) for v in value))
    return env.call(op, fpr, fpr.right, value)


@ufunc.resolver(SqlQueryBuilder, DataType, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, dtype: DataType, value: Any):
    column = env.backend.get_column(env.table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, UUID, str, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, dtype: DataType, value: Any):
    column = env.backend.get_column(env.table, dtype.prop).cast(sa.String)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, Text, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, dtype: Text, value: Any):
    column = get_language_column(env, dtype)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, object, names=COMPARE)
def compare(
    env: SqlQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: DataType,
    value: Any,
):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, Function, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, func: Function, value: Any):
    return _sa_compare(op, func, value)


@ufunc.resolver(SqlQueryBuilder, DataType, list)
def eq(env: SqlQueryBuilder, dtype: DataType, value: List[Any]):
    column = env.backend.get_column(env.table, dtype.prop)
    return column.in_(value)


@ufunc.resolver(SqlQueryBuilder, Text, list)
def eq(env: SqlQueryBuilder, dtype: Text, value: List[Any]):
    column = get_language_column(env, dtype)
    return column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, list)
def eq(env: SqlQueryBuilder, fpr: ForeignProperty, dtype: DataType, value: list):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop)
    return column.in_(value)


@ufunc.resolver(SqlQueryBuilder, DataType, list)
def ne(env: SqlQueryBuilder, dtype: DataType, value: List[Any]):
    column = env.backend.get_column(env.table, dtype.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, Text, list)
def ne(env: SqlQueryBuilder, dtype: Text, value: List[Any]):
    column = get_language_column(env, dtype)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, list)
def ne(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: List[Any],
):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, fpr.right.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, list)
def ne(env: SqlQueryBuilder, fpr: ForeignProperty, value: List[Any]):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, fpr.right.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, object)
def _resolve_unresolved(env: SqlQueryBuilder, value: Any) -> Any:
    if isinstance(value, Unresolved):
        raise ValueError(f"Unresolved value {value!r}.")
    else:
        return value


@ufunc.resolver(SqlQueryBuilder, Bind)
def _resolve_unresolved(env: SqlQueryBuilder, field: Bind) -> sa.Column:
    prop = env.model.flatprops.get(field.name)
    if prop:
        return env.backend.get_column(env.table, prop)
    else:
        raise PropertyNotFound(env.model, property=field.name)


@ufunc.resolver(SqlQueryBuilder, GetAttr)
def _resolve_unresolved(env: SqlQueryBuilder, attr: GetAttr) -> sa.Column:
    fpr = env.call('_resolve_getattr', attr)
    table = env.joins.get_table(env, fpr)
    dtype = fpr.right
    return env.backend.get_column(table, dtype.prop)


@ufunc.resolver(SqlQueryBuilder, Expr, name='and')
def and_(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    args = [
        env.call('_resolve_unresolved', arg)
        for arg in args
        if arg is not None
    ]
    if len(args) > 1:
        return sa.and_(*args)
    elif args:
        return args[0]


@ufunc.resolver(SqlQueryBuilder, Expr, name='or')
def or_(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    args = [
        env.call('_resolve_unresolved', arg)
        for arg in args
        if arg is not None
    ]
    if len(args) > 1:
        return sa.or_(*args)
    elif args:
        return args[0]


@ufunc.resolver(SqlQueryBuilder, Expr, name='list')
def list_(env: SqlQueryBuilder, expr: Expr) -> List[Any]:
    args, kwargs = expr.resolve(env)
    return list(args)


@ufunc.resolver(SqlQueryBuilder)
def count(env: SqlQueryBuilder):
    return sa.func.count()


def _get_property_for_select(
    env: SqlQueryBuilder,
    name: str,
    *,
    nested: bool = False,
) -> Property:
    # TODO: `name` can refer to (in specified order):
    #       - var - a defined variable
    #       - param - a parameter if parametrization is used
    #       - item - an item of a dict or list
    #       - prop - a property
    #       Currently only `prop` is resolved.
    prop = env.model.flatprops.get(name)
    if prop and (
        # Check authorization only for top level properties in select list.
        # XXX: Not sure if nested is the right property to user, probably better
        #      option is to check if this call comes from a prepare context. But
        #      then how prepare context should be defined? Probably resolvers
        #      should be called with a different env class?
        #      tag:resolving_private_properties_in_prepare_context
        nested or
        authorized(env.context, prop, Action.SEARCH)
    ):
        return prop
    else:
        raise PropertyNotFound(env.model, property=name)


@ufunc.resolver(SqlQueryBuilder, Expr)
def select(env: SqlQueryBuilder, expr: Expr):
    keys = [str(k) for k in expr.args]
    args, kwargs = expr.resolve(env)
    args = list(zip(keys, args)) + list(kwargs.items())
    if env.selected is not None:
        raise RuntimeError("`select` was already called.")
    env.selected = {}
    if args:
        for key, arg in args:
            selected = env.call('select', arg)
            if selected is not None:
                if selected.prop is None or selected.prop is not None and not selected.prop.dtype.inherited:
                    env.selected[key] = selected
    else:
        for prop in take(['_id', all], env.model.properties).values():
            if authorized(env.context, prop, Action.GETALL):
                processed = env.call('select', prop)
                if not prop.dtype.inherited or processed.prep is not None:
                    env.selected[prop.place] = processed

    if not (len(args) == 1 and args[0][0] == '_page'):
        if not env.columns:
            raise RuntimeError(
                f"{expr} didn't added anything to select list."
            )


@ufunc.resolver(SqlQueryBuilder, sa.sql.expression.ColumnElement)
def select(env, column):
    return Selected(env.add_column(column))


@ufunc.resolver(SqlQueryBuilder, Bind)
def select(env: SqlQueryBuilder, item: Bind, *, nested: bool = False):
    if item.name == '_page':
        return None
    prop = _get_property_for_select(env, item.name, nested=nested)
    return env.call('select', prop)


@ufunc.resolver(SqlQueryBuilder, str)
def select(env: SqlQueryBuilder, item: str, *, nested: bool = False):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    prop = _get_property_for_select(env, item, nested=nested)
    return env.call('select', prop)


@ufunc.resolver(SqlQueryBuilder, Property)
def select(env: SqlQueryBuilder, prop: Property) -> Selected:
    if prop.place not in env.resolved:
        if isinstance(prop.external, list):
            raise SourceCannotBeList(prop)
        if prop.external.prepare is not NA:
            # If `prepare` formula is given, evaluate formula.
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = process_literal_value(prop.external.prepare)

            # XXX: Maybe interpretation of prepare formula should be done under
            #      a different Env? This way, select resolvers would know when
            #      properties are resolved under a formula context and for
            #      example would be able to decide to not check permissions on
            #      properties.
            #      tag:resolving_private_properties_in_prepare_context
            result = env.call('select', prop.dtype, result)
        elif prop.external and prop.external.name:
            # If prepare is not given, then take value from `source`.
            result = env.call('select', prop.dtype)
        elif prop.is_reserved():
            # Reserved properties never have external source.
            result = env.call('select', prop.dtype)
        elif not prop.dtype.requires_source:
            # Some DataTypes might have children that have source instead of themselves, like: Text, Object
            result = env.call('select', prop.dtype)
        elif prop.dtype.inherited:
            # Some DataTypes might be inherited, or hidden, so we need to go through them in case they can be joined
            result = env.call('select', prop.dtype)
            if not isinstance(result, Selected):
                result = Selected(prop=prop, prep=None)
        else:
            # If `source` is not given, return None.
            result = Selected(prop=prop, prep=None)
        assert isinstance(result, Selected), prop
        env.resolved[prop.place] = result
    return env.resolved[prop.place]


@ufunc.resolver(SqlQueryBuilder, DataType)
def select(env: SqlQueryBuilder, dtype: DataType) -> Selected:
    table = env.backend.get_table(env.model)
    column = env.backend.get_column(table, dtype.prop, select=True)
    return Selected(
        item=env.add_column(column),
        prop=dtype.prop,
    )


@ufunc.resolver(SqlQueryBuilder, Object)
def select(env: SqlQueryBuilder, dtype: Object) -> Selected:
    prep = {}
    for prop in take(dtype.properties).values():
        sel = env.call('select', prop)
        if sel is not None:
            prep[prop.name] = sel
    return Selected(prop=dtype.prop, prep=prep)


@ufunc.resolver(SqlQueryBuilder, Ref)
def select(env: SqlQueryBuilder, dtype: Ref) -> Selected:
    table = env.backend.get_table(env.model)
    column = env.backend.get_column(table, dtype.prop, select=True)

    for prop in dtype.properties.values():
        processed = env.call("select", prop)
        if not prop.dtype.inherited or processed.prep is not None:
            env.selected[prop.place] = processed

    if column is not None and not dtype.inherited:
        return Selected(
            item=env.add_column(column),
            prop=dtype.prop,
        )


@ufunc.resolver(SqlQueryBuilder, Ref, str)
def select(env, dtype: Ref, prop: str):
    table = env.backend.get_table(env.model)
    return Selected(
        env.add_column(
            table.c[dtype.prop.place + '.' + prop]
        )
    )


@ufunc.resolver(SqlQueryBuilder, String)
def select(env, dtype):
    env.call('validate_dtype_for_select', dtype, _gather_selected_properties(env))
    column = env.backend.get_column(env.table, dtype.prop)
    return Selected(env.add_column(column), dtype.prop)


@ufunc.resolver(SqlQueryBuilder, Text)
def select(env, dtype):
    env.call('validate_dtype_for_select', dtype, _gather_selected_properties(env))
    if env.query_params.push:
        result = {
            key: env.call('select', prop)
            for key, prop in dtype.langs.items()
        }
        return Selected(prop=dtype.prop, prep=result)

    if env.query_params.lang:
        for lang in env.query_params.lang:
            if isinstance(lang, Star):
                result = {
                    key: env.call('select', prop)
                    for key, prop in dtype.langs.items()
                }
                return Selected(prop=dtype.prop, prep=result)
            break

        result = {
            key: env.call('select', prop)
            for key, prop in dtype.langs.items() if key in env.query_params.lang
        }
        return Selected(prop=dtype.prop, prep=result)

    default_langs = env.context.get('config').languages
    lang_prop = determine_language_property_for_text(dtype, env.query_params.lang_priority, default_langs)
    return Selected(prop=dtype.prop, prep={
        lang_prop.name: env.call('select', lang_prop)
    })


@ufunc.resolver(SqlQueryBuilder, Denorm)
def select(env, dtype: Denorm):
    ref = dtype.prop.parent
    root_parent = ref
    if isinstance(ref, Property) and isinstance(ref.dtype, Ref):
        fpr = None
        if ref.dtype.inherited:
            parent_list = []
            root_ref_parent = ref
            while root_ref_parent and isinstance(root_ref_parent, Property) and isinstance(root_ref_parent.dtype, Ref):
                parent_list.append(root_ref_parent)
                if not root_ref_parent.dtype.inherited:
                    break
                root_ref_parent = root_ref_parent.parent

            if parent_list:
                parent_list = list(reversed(parent_list))
                root_parent = parent_list.pop(0)
                for parent in parent_list:
                    if parent.place.startswith(f'{root_parent.place}.'):
                        fixed_name = parent.place.replace(f'{root_parent.place}.', '', 1)
                        if fixed_name in root_parent.dtype.model.properties:
                            parent = root_parent.dtype.model.properties[fixed_name]
                    fpr = ForeignProperty(fpr, root_parent.dtype, parent.dtype)
                    root_parent = parent
        fpr = ForeignProperty(fpr, root_parent.dtype, dtype.rel_prop.dtype)
        return env.call("select", fpr)


@ufunc.resolver(SqlQueryBuilder, Page)
def select(env: SqlQueryBuilder, page: Page) -> List[sa.Column]:
    table = env.backend.get_table(env.model)
    return_selected = []
    for item in page.by.values():
        column = env.backend.get_column(table, item.prop, select=True)
        return_selected.append(column)
    return return_selected


@ufunc.resolver(SqlQueryBuilder, PrimaryKey)
def select(
    env: SqlQueryBuilder,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys
    if not pkeys:
        # If primary key is not specified use all properties to uniquely
        # identify row.
        pkeys = take(model.properties).values()

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', prop)
    else:
        result = [
            env.call('select', prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
) -> Selected:
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop, select=True)
    return Selected(
        item=env.add_column(column),
        prop=dtype.prop,
    )


@ufunc.resolver(SqlQueryBuilder, ForeignProperty)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
) -> Selected:
        table = env.joins.get_table(env, fpr)
        right = fpr.right.prop
        column = env.backend.get_column(table, right, select=True)
        return Selected(
            item=env.add_column(column),
            prop=right,
        )


@ufunc.resolver(SqlQueryBuilder, Geometry, Flip)
def select(env: SqlQueryBuilder, dtype: Geometry, func_: Flip):
    table = env.backend.get_table(env.model)

    if dtype.prop.list is None:
        column = env.backend.get_column(table, dtype.prop, select=True)
    else:
        column = env.backend.get_column(table, dtype.prop.list, select=True)

    column = dialect_specific_geometry_flip(env.backend.engine, column)
    return Selected(env.add_column(column), prop=dtype.prop)


@ufunc.resolver(SqlQueryBuilder, Property)
def join_table_on(env: SqlQueryBuilder, prop: Property) -> Any:
    if prop.external.prepare is not NA:
        if isinstance(prop.external.prepare, Expr):
            result = env.resolve(prop.external.prepare)
        else:
            result = process_literal_value(prop.external.prepare)
        return env.call('join_table_on', prop.dtype, result)
    else:
        return env.call('join_table_on', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, DataType)
def join_table_on(env: SqlQueryBuilder, dtype: DataType) -> Tuple[Any]:
    column = env.backend.get_column(env.table, dtype.prop)
    return column


@ufunc.resolver(SqlQueryBuilder, Text)
def join_table_on(env: SqlQueryBuilder, dtype: Text) -> Tuple[Any]:
    column = get_language_column(env, dtype)
    return column


@ufunc.resolver(SqlQueryBuilder, PrimaryKey)
def join_table_on(env: SqlQueryBuilder, dtype: DataType) -> Any:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('join_table_on', prop)
    else:
        result = [
            env.call('join_table_on', prop)
            for prop in pkeys
        ]

    return result


@ufunc.resolver(SqlQueryBuilder, DataType, tuple)
def join_table_on(
    env: SqlQueryBuilder,
    dtype: DataType,
    prep: tuple,
) -> Tuple[Any]:
    return tuple(env.call('join_table_on', v) for v in prep)


@ufunc.resolver(SqlQueryBuilder, Bind)
def join_table_on(env: SqlQueryBuilder, item: Bind):
    prop = env.model.flatprops.get(item.name)
    if not prop or not authorized(env.context, prop, Action.SEARCH):
        raise PropertyNotFound(env.model, property=item.name)
    return env.call('join_table_on', prop)


@ufunc.resolver(SqlQueryBuilder, LiteralProperty)
def join_table_on(env: SqlQueryBuilder, item: LiteralProperty):
    return item.value


@ufunc.resolver(SqlQueryBuilder, DataType, LiteralProperty)
def join_table_on(env: SqlQueryBuilder, dtype: DataType, item: LiteralProperty):
    return env.call('join_table_on', item)


@ufunc.resolver(SqlQueryBuilder, Bind, name='len')
def len_(env: SqlQueryBuilder, bind: Bind):
    prop = env.model.flatprops[bind.name]
    return env.call('len', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, str, name='len')
def len_(env: SqlQueryBuilder, bind: str):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    prop = env.model.flatprops[bind]
    return env.call('len', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, DataType, name='len')
def len_(env: SqlQueryBuilder, dtype: DataType):
    column = env.backend.get_column(env.table, dtype.prop)
    return sa.func.length(column)


@ufunc.resolver(SqlQueryBuilder, Expr)
def sort(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    env.sort = []
    for key in args:
        result = env.call('sort', key)
        if isinstance(result, (list, set, tuple)):
            env.sort += result
        else:
            env.sort.append(result)


@ufunc.resolver(SqlQueryBuilder, DataType)
def sort(env, dtype):
    return env.call('asc', dtype)


@ufunc.resolver(SqlQueryBuilder, Bind)
def sort(env, field):
    prop = env.model.get_from_flatprops(field.name)
    return env.call('asc', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, Negative)
def sort(env, field):
    prop = env.model.get_from_flatprops(field.name)
    return env.call('desc', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, DataType)
def asc(env, dtype):
    column = env.backend.get_column(env.table, dtype.prop)
    return dialect_specific_asc(env.backend.engine, column)


@ufunc.resolver(SqlQueryBuilder, Text)
def asc(env, dtype):
    column = get_language_column(env, dtype)
    return dialect_specific_asc(env.backend.engine, column)


@ufunc.resolver(SqlQueryBuilder, DataType)
def desc(env, dtype):
    column = env.backend.get_column(env.table, dtype.prop)
    return dialect_specific_desc(env.backend.engine, column)


@ufunc.resolver(SqlQueryBuilder, Text)
def desc(env, dtype):
    column = get_language_column(env, dtype)
    return dialect_specific_desc(env.backend.engine, column)


@ufunc.resolver(SqlQueryBuilder, DataType)
def negative(env: SqlQueryBuilder, dtype: DataType):
    return Negative(dtype.prop.place)


@ufunc.resolver(SqlQueryBuilder, GetAttr)
def negative(env: SqlQueryBuilder, attr: GetAttr):
    resolved = env.call('_resolve_getattr', attr)
    return env.call('negative', resolved)


@ufunc.resolver(SqlQueryBuilder, String)
def negative(env: SqlQueryBuilder, dtype: String):
    if dtype.prop.parent and isinstance(dtype.prop.parent.dtype, Text):
        return Negative(dtype.prop.place.replace('.', '@'))
    return Negative(dtype.prop.place)


@ufunc.resolver(SqlQueryBuilder, int)
def limit(env: SqlQueryBuilder, n: int):
    env.limit = n


@ufunc.resolver(SqlQueryBuilder, int)
def offset(env: SqlQueryBuilder, n: int):
    env.offset = n


@ufunc.resolver(SqlQueryBuilder, Expr)
def file(env: SqlQueryBuilder, expr: Expr) -> Expr:
    args, kwargs = expr.resolve(env)
    return Expr(
        'file',
        name=env.call('select', kwargs['name'], nested=True),
        content=env.call('select', kwargs['content'], nested=True),
    )


@overload
@ufunc.resolver(SqlQueryBuilder)
def cast(env: SqlQueryBuilder) -> Expr:
    return Expr('cast')


@overload
@ufunc.resolver(SqlQueryBuilder, Bind, Bind)
def point(env: SqlQueryBuilder, x: Bind, y: Bind) -> Expr:
    return Expr(
        'point',
        env.call('select', x, nested=True),
        env.call('select', y, nested=True),
    )


@ufunc.resolver(SqlQueryBuilder)
def distinct(env: SqlQueryBuilder):
    if env.model and env.model.external and not env.model.external.unknown_primary_key:
        extracted_columns = [env.backend.get_column(env.table, prop) for prop in env.model.external.pkeys if prop.external]
        if env.group_by is None:
            env.group_by = extracted_columns
        else:
            for column in extracted_columns:
                if column not in env.group_by:
                    env.group_by.append(column)
    else:
        env.distinct = True


@ufunc.resolver(SqlQueryBuilder, Expr)
def swap(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    return Expr('swap', *args, **kwargs)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, PrimaryKey)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: PrimaryKey,
) -> Selected:
    super_ = ufunc.resolver[env, fpr, dtype]
    return super_(env, fpr, dtype)


@ufunc.resolver(SqlQueryBuilder, Geometry)
def flip(env: SqlQueryBuilder, dtype: Geometry):
    if contains_geometry_flip_function(env.backend.engine):
        return Flip(dtype)

    # Returning expr means, that it will be passed to ResultBuilder to handle it
    return Expr('flip')

